"""This file provides the ability to combine vulnerabilities from multiple sources into a single model.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/vulnerability_combiner.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from typing_extensions import deprecated


if TYPE_CHECKING:
    from hoppr import Vulnerability


def _get_all_ids_from_vulnerability(from_vulnerability: Vulnerability) -> list[str]:
    ids: list[str] = [from_vulnerability.id] if from_vulnerability.id else []

    if from_vulnerability.references is not None:
        ids.extend(reference.id for reference in from_vulnerability.references)

    return list(set(ids))


def _does_id_exist_in_both(existing_id_list: list[str], new_id_list: list[str]) -> bool:
    return len([value for value in existing_id_list if value in new_id_list]) > 0


def _get_matching_vulnerability_from_list(
    combined_list: list[Vulnerability], new_finding: Vulnerability
) -> Vulnerability | None:
    new_finding_ids = _get_all_ids_from_vulnerability(new_finding)
    for vuln in combined_list:
        existing_finding_ids = _get_all_ids_from_vulnerability(vuln)

        if _does_id_exist_in_both(existing_finding_ids, new_finding_ids):
            return vuln

    return None


def merge_scanner_vulnerabilities(
    scanner_vulns: dict[str, list[Vulnerability]],
) -> list[Vulnerability]:
    """Intelligently combines the lists of vulnerabilities found by different scanners."""
    combined_result: list[Vulnerability] = []

    for vulns in scanner_vulns.values():
        for vulnerability in vulns:
            if existing := _get_matching_vulnerability_from_list(combined_result, vulnerability):
                existing.merge(vulnerability)
            else:
                combined_result.append(vulnerability)

    return combined_result


@deprecated(
    "vulnerability_combiner.combine_vulnerabilities is deprecated, use vulnerability_combiner.merge_scanner_vulnerabilities"
)
def combine_vulnerabilities(
    vulnerabilities_list: list[dict[str, list[Vulnerability]]],
) -> dict[str, list[Vulnerability]]:
    """Intelligently combines the lists of vulnerabilities found by different scanners."""
    combined_result: dict[str, list[Vulnerability]] = {}

    for result in vulnerabilities_list:
        for purl in result:
            tool_vulnerabilities = result.get(purl)

            if purl not in combined_result:
                combined_result[purl] = []

            for vulnerability in tool_vulnerabilities or []:
                if existing := _get_matching_vulnerability_from_list(combined_result[purl], vulnerability):
                    existing.merge(vulnerability)
                else:
                    combined_result[purl].append(vulnerability)

    return combined_result



"""Utility functions for working with vulnerabilities and SBOMs.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/utils.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import json
import os
import stat
import tempfile

from pathlib import Path
from typing import TYPE_CHECKING, Any

import requests
import rich
import typer

from hoppr import Component, ComponentType, Sbom, Vulnerability, cdx
from hoppr.utils import dedup_list, get_package_url
from rapidfuzz import fuzz
from requests import Response
from typing_extensions import deprecated


if TYPE_CHECKING:
    from packageurl import PackageURL


FUZZY_MATCH_THRESHOLD = 85


def purl_check(first_purl_str: str, second_purl_str: str) -> bool:
    """Determine if two purls are equivalent. Returns true if all properties of the PURL match."""
    if not (first_purl_str and second_purl_str):
        return False

    self_purl = get_package_url(first_purl_str)
    other_purl = get_package_url(second_purl_str)

    first_qualifiers = {}
    second_qualifiers = {}
    match self_purl.qualifiers, other_purl.qualifiers:
        case str(first), str(second):
            first_qualifiers = {"encoded": first}
            second_qualifiers = {"encoded": second}
        case str(first), dict(second):
            first_qualifiers = {"encoded": first}
            second_qualifiers = second
        case dict(first), str(second):
            first_qualifiers = first
            second_qualifiers = {"encoded": second}
        case dict(first), dict(second):
            first_qualifiers = first
            second_qualifiers = second

    qual_keys = dedup_list(first_qualifiers | second_qualifiers)

    return all(
        [
            self_purl.name == other_purl.name,
            self_purl.type == other_purl.type,
            self_purl.namespace == other_purl.namespace,
            str(self_purl.version).removeprefix("v") == str(other_purl.version).removeprefix("v"),
            self_purl.subpath == other_purl.subpath,
            *[qualifier_match(key, first_qualifiers, second_qualifiers) for key in qual_keys],
        ]
    )


def qualifier_match(key: str, first_qualifiers: dict[str, str], second_qualifiers: dict[str, str]) -> bool:
    """Determines if two groups of PURL qualifiers contain the same entries using a fuzzy match."""
    # Compare only if both purls have a value for specified qualifier
    if key not in first_qualifiers or key not in second_qualifiers:
        return True

    return fuzz.ratio(first_qualifiers.get(key, ""), second_qualifiers.get(key, "")) > FUZZY_MATCH_THRESHOLD


def _add_vulnerability(key: str, vuln_list: dict[str, Vulnerability], new_vuln: Vulnerability):
    vuln_list.setdefault(key, Vulnerability())
    vuln_list[key].merge(new_vuln)


def unsupported_purl_feedback(scanner_name: str, supported_types: list[str], purl_list: list[PackageURL]) -> None:
    """Determine if there are purls that the scanner does not support, report findings."""
    purl_types = list(dict.fromkeys(purl.type for purl in purl_list))

    for purl_type in filter(lambda type_: type_ not in supported_types, purl_types):
        rich.print(
            f"[yellow]WARNING: {scanner_name} -- does not support purls of type {purl_type}, components may be missed.",
        )


def convert_xml_to_json(file_path: Path) -> Path:
    """Function to convert a xml file to json format."""
    typer.echo("xml format detected, attempt to convert with cyclonedx tools")

    # Default to the path specified in the hoppr-cop docker file or define the local filename to save data
    docker_image_path = Path("/usr/local/bin/cyclone-dx")
    cyclone_dx_path = docker_image_path if docker_image_path.exists() else Path(tempfile.gettempdir()) / "cyclonedx"
    if not cyclone_dx_path.exists():
        typer.echo("cyclonedx tools not found Attempting to download")
        if url := os.getenv("CYCLONE_INSTALL_URL"):
            # Make http request for remote file datae
            data = requests.get(url, timeout=60)

            # Save file data to local copy
            cyclone_dx_path.write_bytes(data.content)
        else:
            msg = typer.style(
                "In order to support xml SBOMs, you must set 'CYCLONE_INSTALL_URL' to "
                "the correct release of cyclone-dx cli. https://github.com/CycloneDX/cyclonedx-cli/releases",
                fg=typer.colors.RED,
            )
            typer.echo(msg)
            raise typer.Exit(code=1)

    cyclone_dx_path.chmod(stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

    os.system(
        f"{cyclone_dx_path.absolute()} convert --input-file {file_path} \
            --output-file {tempfile.gettempdir()}/{file_path.name}.json --output-format json",
    )
    return Path(f"{tempfile.gettempdir()}/{file_path.name}.json")


def parse_sbom(sbom_file: Path) -> Sbom:
    """Parses a Software Bill of Materials File."""
    if sbom_file.is_file() and sbom_file.exists():
        typer.echo(f"processing {sbom_file}")

        if sbom_file.suffix == ".xml":
            sbom_file = convert_xml_to_json(sbom_file)

        sbom_json_object = json.loads(sbom_file.read_text(encoding="utf-8"))
        sbom = create_sbom_object(sbom_json_object, str(sbom_file))
    else:
        typer.secho(f"{sbom_file!s} is not a file", fg=typer.colors.RED)
        raise typer.Exit

    return sbom


def parse_sbom_json_string(sbom_file_json: str, sbom_info: str) -> Sbom:
    """Parses a Software Bill of Materials JSON String."""
    sbom_json_object = json.loads(sbom_file_json)
    return create_sbom_object(sbom_json_object, sbom_info)


def create_sbom_object(sbom_json_object: dict[str, Any], sbom_info: str) -> Sbom:
    """Creates a Software Bill of Materials Object."""
    spec_version = sbom_json_object.get("specVersion", "")
    sbom_json_object.pop("$schema", None)

    if spec_version not in {"1.2", "1.3", "1.4", "1.5", "1.6"}:
        typer.secho(f"{sbom_info} is an unknown spec version ({spec_version})")
        raise typer.Exit

    return Sbom.parse_obj(sbom_json_object)


def get_vulnerability_source(vulnerability_id: str) -> cdx.VulnerabilitySource | None:
    """Generate the source for a vulnerability based on a given ID."""
    match vulnerability_id:
        case cve if cve.startswith("CVE-"):
            return cdx.VulnerabilitySource(
                name="NVD",
                url=f"https://nvd.nist.gov/vuln/detail/{vulnerability_id}",
            )
        case gh if gh.startswith(("GHSA", "GMS")):
            return cdx.VulnerabilitySource(
                name="Github Advisories",
                url=f"https://github.com/advisories/{vulnerability_id}",
            )
        case oss if oss.startswith("sonatype"):
            return cdx.VulnerabilitySource(
                name="OSS Index",
                url=f"https://ossindex.sonatype.org/vulnerability/{vulnerability_id}",
            )
        case _:
            return None


def get_advisories_from_urls(urls: list[str]) -> list[cdx.Advisory]:
    """Generates a list of advisories for the given set of urls."""
    urls = list(set(urls))
    return [cdx.Advisory(url=x) for x in urls]


def get_references_from_ids(ids: list[str], primary_id: str) -> list[cdx.Reference]:
    """Builds a list of Reference objects to the given vulnerability IDs."""
    references = []

    for ident in list(set(ids)):
        if ident != primary_id and (source := get_vulnerability_source(ident)):
            references.append(cdx.Reference(id=ident, source=source))

    return references


def build_bom_dict_from_purls(purls: list[PackageURL]) -> dict[str, Any]:
    """Create SBOM dictionary from PackageURL list."""
    sbom = Sbom(
        components=[
            Component(
                type=ComponentType.LIBRARY,
                name=purl.name if purl.type == "npm" else f"{purl.namespace}/{purl.name}",
                version=purl.version,
                purl=purl.to_string(),
                group=purl.namespace,
                bom_ref=purl.to_string(),
                description="test",
                author="test",
                externalReferences=[],
            )
            for purl in purls
        ]
    )

    return sbom.dict()


def create_bom_from_purl_list(purls: list[str]) -> Sbom:
    """Creates a skeleton SBOM from a list of purls."""
    return Sbom.parse_obj(build_bom_dict_from_purls([get_package_url(purl) for purl in purls]))


@deprecated("utils.build_bom_from_purls is deprecated, use utils.create_bom_from_purl_list")
def build_bom_from_purls(purls: list[PackageURL]) -> Sbom:
    """Creates a skeleton SBOM from a list of purls."""
    return Sbom.parse_obj(build_bom_dict_from_purls(purls))


def api_query(url: str, proxies: dict[str, str] | None = None) -> Response:
    """Load response from API query into JSON."""
    return requests.get(
        url=url,
        allow_redirects=True,
        proxies=proxies,
        stream=True,
        verify=True,
        timeout=60,
    )



"""A Vulnerability Scanner that combines results from all configured scanners.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/combined/cli.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

from pathlib import Path
from platform import python_version

import rich
import typer

from rich.progress import Progress, SpinnerColumn, TextColumn
from typer import Typer

import hopprcop

from hopprcop.combined.combined_scanner import CombinedScanner
from hopprcop.enhancements.epss_enhancer import EpssEnhancer
from hopprcop.gemnasium.gemnasium_scanner import GemnasiumScanner
from hopprcop.grype.grype_scanner import GrypeScanner
from hopprcop.ossindex.oss_index_scanner import OSSIndexScanner
from hopprcop.reporting import Reporting
from hopprcop.reporting.models import CycloneDxRenderOptions, ReportFormat
from hopprcop.trivy.trivy_scanner import TrivyScanner
from hopprcop.utils import parse_sbom


app = Typer(
    name="hoppr-cop",
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Policing Your SBOM Vulnerabilities",
    no_args_is_help=False,
    pretty_exceptions_show_locals=False,
    rich_markup_mode="markdown",
)


def _output_dir(output_dir: Path | None) -> Path:
    """Use `Path.cwd()` as output directory if not provided."""
    return output_dir or Path.cwd()


def _version(version: bool) -> None:
    if version:
        rich.print(f"[green]HopprCop Version       [/] : {hopprcop.__version__}")
        rich.print(f"[green]Python Version         [/] : {python_version()}")
        raise typer.Exit(code=0)


def _formats(formats: list[ReportFormat] | None) -> list[str]:
    """Use `[ReportFormat.TABLE]` as report format if not provided."""
    return [format_.value for format_ in formats or [ReportFormat.TABLE]]


@app.callback(invoke_without_command=True)
def vulnerability_report(
    bom: str = typer.Argument(
        None,
        help="Path to a CycloneDX SBOM",
        show_default=False,
    ),
    formats: list[ReportFormat] = typer.Option(
        None,
        "--format",
        "-f",
        callback=_formats,
        encoding="utf-8",
        help='The report formats to generate [default: ["table"]]',
        show_default=False,
    ),
    output_dir: Path = typer.Option(
        None,
        "--output-dir",
        "-d",
        callback=_output_dir,
        help="The directory where reports will be written [default: .]",
        show_default=False,
    ),
    base_report_name: str = typer.Option(
        None,
        "--base-report-name",
        "-b",
        help='The base name supplied for the generated reports [default: "hoppr-cop-report"]',
        show_default=False,
    ),
    os_distro: str = typer.Option(
        None,
        "--os-distro",
        "-o",
        envvar="OS_DISTRIBUTION",
        help=(
            "The operating system distribution; this is important "
            "to ensure accurate reporting of OS vulnerabilities from grype. "
            "Examples include rhel:8.6 or rocky:9"
        ),
        show_default=False,
    ),
    offline: bool = typer.Option(
        False,
        "--offline",
        help="Mode to allow for offline scans",
        rich_help_panel="Experimental Options",
        show_default=False,
    ),
    download_dbs: bool = typer.Option(
        False,
        "--download-dbs",
        help="Mode to allow for downloading vulnerabilty databases",
        show_default=False,
    ),
    trace: bool = typer.Option(
        False,
        help="Print traceback information on unhandled error [default: no-trace]",
        show_default=False,
    ),
    assessment: str = typer.Option(
        None,
        help="Directory path to your analysis.assessment.yml [default: None]",
        show_default=False,
    ),
    epss: bool = typer.Option(
        False,
        "--epss",
        help="Enable EPSS score enhancement",
        show_default=False,
    ),
    version: bool = typer.Option(
        False,
        "--version",
        callback=_version,
        is_eager=True,
        help="Show HopprCop version information",
    ),
    keep_serial_numbers: bool = typer.Option(False, help="Do not make a new serial number when enhancing SBOMs"),
    increment_version: bool = typer.Option(
        False,
        help=("Increment the version of the SBOM when enhancing. Only used if --keep-serial-numbers is also specifed"),
    ),
):
    """Generates vulnerability reports based on the specified SBOM and formats."""
    try:
        if not (bom or download_dbs):
            raise typer.BadParameter("Either a path to an SBOM file or the --download-dbs option is required")

        elif download_dbs:
            combined = CombinedScanner()
            combined.set_scanners(
                [
                    GrypeScanner(),
                    TrivyScanner(),
                    GemnasiumScanner(),
                ]
            )

            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
            ) as progress:
                progress.add_task(description="Fetching vulnerability databases...", total=None)
                combined.get_vulnerability_dbs()
                rich.print("Successfully downloaded vulnerability databases")
                typer.Exit(code=0)

        else:
            bom_path = Path(bom)
            if not bom_path.exists():
                rich.print(f"[red] {bom} does not exist")
                raise typer.Exit(code=1)

            if base_report_name is None:
                base_report_name = bom_path.stem

            reporting = Reporting(output_dir, base_report_name)
            combined = CombinedScanner()

            if assessment:
                combined.set_assessment_path(assessment)

            combined.set_scanners(
                [
                    GrypeScanner(offline_mode=offline, grype_os_distro=os_distro),
                    TrivyScanner(offline_mode=offline, trivy_os_distro=os_distro),
                    OSSIndexScanner(offline_mode=offline),
                    GemnasiumScanner(offline_mode=offline),
                ]
            )

            combined.set_enhancers([EpssEnhancer(offline_mode=offline, enabled=epss)])

            parsed_bom = parse_sbom(bom_path)

            render_options = CycloneDxRenderOptions(
                keep_serial_numbers=keep_serial_numbers, increment_version=increment_version
            )

            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                transient=True,
            ) as progress:
                progress.add_task(description="Fetching vulnerabilities...", total=None)
                results = combined.get_vulnerabilities_for_sbom(parsed_bom)

            reporting.generate_vulnerability_reports(formats, results, parsed_bom, options=render_options)
    except Exception as exc:
        if trace:
            rich.get_console().print_exception(show_locals=False)

        rich.print(f"[red]unexpected error: {exc}")
        raise typer.Exit(code=1) from exc



"""Provides vulnerability reporting.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2023 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/reporting/__init__.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import re
import uuid

from collections import defaultdict
from collections.abc import Mapping
from datetime import datetime, timezone
from pathlib import Path
from typing import overload

import jinja2
import rich

from hoppr import Affect, Component, Metadata, Sbom, Tools, Vulnerability, cdx
from hoppr.utils import dedup_list, get_package_url
from rich.box import Box
from rich.table import Table
from typing_extensions import deprecated

import hopprcop

from hopprcop.reporting.gitlab.models import (
    Analyzer,
    CvssVectors,
    CvssVectors3x,
    Dependency,
    DependencyFile,
    Identifier,
    Link,
    Location,
    NamedFieldTextItem,
    Package,
    ReportFormatForGitlabDependencyScanning as GitlabReport,
    Scan,
    Scanner,
    Vendor,
    Vulnerability as GitlabVulnerability,
)
from hopprcop.reporting.models import CycloneDxRenderOptions, ReportFormat


class Reporting:
    """Generates reports in multiple formats from a list of vulnerabilities."""

    output_path: Path
    base_name: str

    def __init__(self, output_path: Path, base_name: str):
        self.output_path = output_path
        self.base_name = base_name

    @overload
    def generate_vulnerability_reports(
        self,
        formats: list[ReportFormat],
        vulnerabilities: Mapping[str, list[Vulnerability]],
        bom: Sbom | None = None,
        options: CycloneDxRenderOptions | None = None,
    ): ...  # pragma: no cover

    @overload
    def generate_vulnerability_reports(
        self,
        formats: list[ReportFormat],
        vulnerabilities: list[Vulnerability],
        bom: Sbom | None = None,
        options: CycloneDxRenderOptions | None = None,
    ): ...  # pragma: no cover

    def generate_vulnerability_reports(
        self,
        formats: list[ReportFormat],
        vulnerabilities: list[Vulnerability] | Mapping[str, list[Vulnerability]],
        bom: Sbom | None = None,
        options: CycloneDxRenderOptions | None = None,
    ):
        """Generates various vulnerability reports based on specified formats."""
        vulns_to_report: list[Vulnerability] = []
        if isinstance(vulnerabilities, Mapping):
            vulns_to_report = self._add_purl_as_bom_ref_and_flatten(vulnerabilities)
        else:
            vulnerabilities.sort(key=self._get_score, reverse=True)
            vulns_to_report = vulnerabilities

        if ReportFormat.CYCLONE_DX in formats and bom is not None:
            self._generate_cyclonedx_report(bom, vulns_to_report, options)

        if ReportFormat.LINKED_VEX in formats and bom is not None:
            self._generate_linked_report(bom, vulns_to_report)

        if ReportFormat.GITLAB in formats:
            self._generate_gitlab_vulnerability_report(vulns_to_report)

        if ReportFormat.HTML in formats:
            self._generate_html_report(vulns_to_report)

        if ReportFormat.TABLE in formats:
            self._generate_table_report(vulns_to_report)

    @deprecated("Reporting._add_purl_as_bom_ref_and_flatten is deprecated and should no longer be necessary")
    def _add_purl_as_bom_ref_and_flatten(
        self, vulnerabilities: Mapping[str, list[Vulnerability]]
    ) -> list[Vulnerability]:
        flattened_vulnerabilities: list[Vulnerability] = []

        for purl, vuln_list in vulnerabilities.items():
            for vuln in vuln_list or []:
                vuln.affects = vuln.affects or []
                if not any(affect for affect in vuln.affects if affect.ref == purl):
                    vuln.affects.append(Affect(ref=purl))
                flattened_vulnerabilities.append(vuln)

        flattened_vulnerabilities.sort(key=self._get_score, reverse=True)
        return flattened_vulnerabilities

    def _set_hopprcop_as_tool(self, bom: Sbom):
        if bom.metadata and bom.metadata.tools:
            bom.metadata.tools.components = dedup_list(
                [
                    Component(
                        type=cdx.Type("application"),
                        name="hoppr-cop",
                        version=hopprcop.__version__,
                        bom_ref=f"pkg:pypi/hoppr-cop@{hopprcop.__version__}",
                        purl=f"pkg:pypi/hoppr-cop@{hopprcop.__version__}",
                        scope=cdx.Scope.EXCLUDED,
                    ),
                    *(bom.metadata.tools.components or []),
                ]
            )
        else:
            bom.metadata = Metadata(
                timestamp=datetime.now(timezone.utc),
                tools=Tools(
                    components=[
                        Component(
                            type=cdx.Type("application"),
                            name="hoppr-cop",
                            version=hopprcop.__version__,
                            bom_ref=f"pkg:pypi/hoppr-cop@{hopprcop.__version__}",
                            purl=f"pkg:pypi/hoppr-cop@{hopprcop.__version__}",
                            scope=cdx.Scope.EXCLUDED,
                        ),
                    ]
                ),
            )

    def _generate_cyclonedx_report(
        self, bom: Sbom, vulnerabilities: list[Vulnerability], options: CycloneDxRenderOptions | None = None
    ):
        """Updates the Software Bill of Materials (SBOM) file with the vulnerabilities found during scanning."""
        self.output_path.mkdir(parents=True, exist_ok=True)
        # Copy the bom so other reports are not affected by changes
        embedded_bom = bom.copy(deep=True)

        # Keep serial number if requested, otherwise make a new one
        if (options is not None and not options.keep_serial_numbers) or options is None:
            embedded_bom.serialNumber = uuid.uuid4().urn
            embedded_bom.version = 1

        # Only need to increment version if we are using the old serial number
        elif options.increment_version:
            embedded_bom.version = embedded_bom.version + 1 if embedded_bom.version is not None else 1
        self._set_hopprcop_as_tool(bom)
        embedded_bom.vulnerabilities = vulnerabilities

        (self.output_path / f"{Path(self.base_name).name}-enhanced.json").write_text(
            embedded_bom.json(indent=2), encoding="utf-8"
        )

    def _generate_linked_report(self, bom: Sbom, vulnerabilities: list[Vulnerability]):
        """Creates a Cyclone DX compliant Software Bill of Materials (SBOM) containing the vulnerabilities found during scanning that is linked to the original SBOM."""
        *_, bom_serial_number = (bom.serialNumber or uuid.uuid4().urn).split(":")

        linked_vulns = [vuln.copy(deep=True) for vuln in vulnerabilities]

        for vuln in linked_vulns:
            vuln.affects = [
                Affect.parse_obj(
                    {
                        "ref": f"urn:cdx:{bom_serial_number}/{bom.version}#{affect.ref.__str__()}",
                        "versions": affect.versions,
                    }
                )
                for affect in vuln.affects
            ]

        vex_bom = Sbom(vulnerabilities=linked_vulns)
        vex_bom.serialNumber = uuid.uuid4().urn
        self._set_hopprcop_as_tool(vex_bom)
        (self.output_path / f"{Path(self.base_name).name}-vex.json").write_text(
            vex_bom.json(indent=2), encoding="utf-8"
        )

    def _get_score(self, vuln_to_score: Vulnerability) -> float:
        """Return best score of specified Vulnerability."""
        best_rating = self._get_best_rating(vuln_to_score.ratings)

        if best_rating is None or best_rating.score is None:
            return 0.0

        return best_rating.score

    def _generate_table_report(self, vulnerabilities: list[Vulnerability]):
        """Creates a Table view containing the vulnerabilities found during scanning, then prints to the terminal."""
        no_border, underline = "    \n", " -  \n"
        simple_box = Box(box=f"{no_border * 2}{underline}{no_border * 5}", ascii=True)

        table = Table("type", "name", "version", "id", "severity", "found by", box=simple_box)

        for finding in self._get_fields_from_vulnerabilities(vulnerabilities):
            table.add_row(*finding)

        rich.print(table)

    def _get_fields_from_vulnerabilities(self, vulnerabilities: list[Vulnerability]) -> list[list[str | None]]:
        findings = []

        def get_fields(vuln: Vulnerability) -> list[list[str | None]]:
            vuln.tools = vuln.tools or Tools()
            vuln_tools = [tool for tool in vuln.tools.components if vuln.tools and vuln.tools.components]
            tools = [f"{(tool.supplier or cdx.OrganizationalEntity()).name} {tool.name}" for tool in vuln_tools]

            severity = self._get_severity(vuln.ratings)

            if severity == "critical":
                severity = "[red]critical[/]"
            elif severity == "high":
                severity = "[bright_yellow]high[/]"

            fields: dict[str, list[str | None]] = {}

            if not vulnerability.affects:
                return [[]]

            for affect in vulnerability.affects:
                purl = get_package_url(affect.ref)
                key = f"{purl.type}/{purl.name}:{purl.version}"
                if key not in fields:
                    fields[key] = [purl.type, purl.name, purl.version, vuln.id, severity, " | ".join(tools)]

            return list(fields.values())

        for vulnerability in vulnerabilities:
            findings.extend(get_fields(vulnerability))

        return findings

    def _copy_assets(self):
        assets_dir = Path(__file__).parent / "templates" / "assets"
        assets = ["vulnerabilities.css"]

        output_path = self.output_path / "assets"
        output_path.mkdir(exist_ok=True, parents=True)

        for asset in assets:
            template_data = (assets_dir / asset).read_text(encoding="utf-8")
            (output_path / asset).write_text(template_data)

    def _generate_html_report(self, combined_list: list[Vulnerability]):
        self.output_path.mkdir(parents=True, exist_ok=True)
        output_path = self.output_path / f"{Path(self.base_name).name}-vulnerabilities.html"

        env = jinja2.Environment(loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates"))
        env.filters["severity"] = self._get_severity
        template = env.get_template("vulnerabilities.html")

        self._copy_assets()

        severity_classes = {
            "critical": "bg-red-100 rounded-lg py-5 px-6 mb-4 text-base text-red-700 mb-3",
            "high": "bg-yellow-100 rounded-lg py-5 px-6 mb-4 text-base text-yellow-700 mb-3",
            "medium": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "info": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "low": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "unknown": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
            "none": "bg-gray-50 rounded-lg py-5 px-6 mb-4 text-base text-gray-500 mb-3",
        }

        result = template.render(
            {"findings": combined_list, "severity_classes": severity_classes, "base_name": self.base_name},
        )

        output_path.write_text(result, encoding="utf-8")

        self._generate_vuln_detail_reports(combined_list, severity_classes)

    def _generate_vuln_detail_reports(self, vulnerabilities: list[Vulnerability], severity_classes: dict[str, str]):
        output_path = self.output_path / f"{Path(self.base_name).name}-details"
        output_path.mkdir(exist_ok=True, parents=True)

        env = jinja2.Environment(loader=jinja2.FileSystemLoader(Path(__file__).parent / "templates"))
        env.filters["featured_link"] = self._get_featured_link
        template = env.get_template("vulnerability_details.html")

        for vuln in vulnerabilities:
            if not vuln.affects:
                continue

            # Need to print all the affected purls not just the first
            # for affect in vuln.affects:
            purls = [get_package_url(affect.ref) for affect in vuln.affects if vuln.affects]
            result = template.render(
                {
                    "components": [
                        {
                            "type": purl.type,
                            "namespace": purl.namespace,
                            "name": purl.name,
                            "version": purl.version,
                            "purl": purl.to_string(),
                        }
                        for purl in purls
                    ],
                    "vulnerability": vuln,
                    "severity_classes": severity_classes,
                    "base_name": self.base_name,
                }
            )

            (output_path / f"{vuln.id}.html").write_text(result, encoding="utf-8")

    def _generate_gitlab_vulnerability_report(self, vulnerabilities: list[Vulnerability]):
        """Renders the vulnerabilities report for gitlab."""
        # Note the JSON schema for this report can be found at
        # https://gitlab.com/gitlab-org/security-products/security-report-schemas/-/blob/master/dist/dependency-scanning-report-format.json

        report = GitlabReport(
            scan=Scan(
                analyzer=Analyzer(
                    id="hoppr-cop", name="Hoppr Cop", version=hopprcop.__version__, vendor=Vendor(name="Hoppr")
                ),
                end_time=datetime.now().replace(microsecond=0).isoformat(),
                scanner=Scanner(
                    id="hoppr-cop", name="Hoppr Cop", version=hopprcop.__version__, vendor=Vendor(name="Hoppr")
                ),
                start_time=datetime.now().replace(microsecond=0).isoformat(),
                status="success",
                type="dependency_scanning",
            ),
            version="15.0.7",
            vulnerabilities=[],
            remediations=[],
            dependency_files=[],
        )

        dependencies_by_format: dict[str, list[Dependency]] = defaultdict(list, {})

        purls = [affect.ref for vuln in vulnerabilities for affect in vuln.affects if vuln.affects]

        self.output_path.mkdir(parents=True, exist_ok=True)
        output_path = self.output_path / "gl-dependency-scanning-report.json"

        for purl_str in purls:
            purl = get_package_url(purl_str)
            dependencies_by_format[purl.type].append(
                Dependency(package=Package(name=purl.name), version=str(purl.version))
            )

        for vuln in vulnerabilities:
            report.vulnerabilities.extend(self._generate_gitlab_row(vuln))

        for repo_format in dependencies_by_format:
            report.dependency_files.append(
                DependencyFile(
                    package_manager=repo_format,
                    path="cyclonedx.bom",
                    dependencies=dependencies_by_format[repo_format],
                )
            )

        output_path.write_text(data=report.json(exclude_none=True, indent=2), encoding="utf-8")

    @staticmethod
    def _get_featured_link(advisories: list[cdx.Advisory] | None) -> str | None:
        if advisories is not None:
            for adv in advisories:
                url = "" if adv.url is None else adv.url
                if "https://snyk.io/" in url:
                    return url
        return None

    def _get_severity(self, ratings: list[cdx.Rating] | None) -> str:
        best_rating = self._get_best_rating(ratings)

        return str(best_rating.severity if best_rating else "none")

    @staticmethod
    def _get_best_rating(ratings: list[cdx.Rating] | None) -> cdx.Rating | None:
        ratings = ratings or []
        default_rating = ratings[0] if ratings else None

        methods = [str(rating.method) if rating.method else "none" for rating in ratings]

        preferred_method = None
        if "CVSSv31" in methods:
            preferred_method = "CVSSv31"
        elif "CVSSv3" in methods:
            preferred_method = "CVSSv3"
        elif "CVSSv2" in methods:
            preferred_method = "CVSSv2"

        return next((rating for rating in ratings if str(rating.method) == preferred_method), default_rating)

    def _get_cvss_vectors(self, ratings: list[cdx.Rating] | None) -> list[CvssVectors3x | CvssVectors]:
        result: list[CvssVectors3x | CvssVectors] = []
        for rating in ratings or []:
            match rating.method:
                case "CVSSv31" | "CVSSv3":
                    result.append(
                        CvssVectors3x(vendor=rating.source.name if rating.source else "unknown", vector=rating.vector)
                    )
                case "CVSSv2":
                    result.append(
                        CvssVectors(vendor=rating.source.name if rating.source else "unknown", vector=rating.vector)
                    )

        return result

    def _generate_gitlab_row(self, vuln: Vulnerability) -> list[GitlabVulnerability]:
        """Generates a report row."""
        # Ensure `affects` and `tools` are non-empty lists
        if not (vuln.id and vuln.affects and vuln.tools):
            return []

        purls = [get_package_url(affect.ref) for affect in vuln.affects if vuln.affects]
        severity = self._get_severity(vuln.ratings).title()

        gitlab_vulns: list[GitlabVulnerability] = [
            GitlabVulnerability(
                id=str(uuid.uuid4()),
                name=vuln.id,
                description=vuln.description,
                severity=severity if severity != "None" else "Info",  # type: ignore[arg-type]
                solution=vuln.recommendation or "",
                identifiers=[Identifier(type=vuln.id.split("-")[0].lower(), name=vuln.id, value=vuln.id)],
                cvss_vectors=self._get_cvss_vectors(vuln.ratings) or None,
                links=[
                    Link(url=advisory.url)
                    for advisory in vuln.advisories or []
                    if re.search("^(https?|ftp)://.+", advisory.url)
                ],
                location=Location(
                    file="cyclonedx.bom",
                    dependency=Dependency(
                        package=Package(name=purl.name),
                        version=purl.version or "",
                    ),
                ),
                details={
                    "vulnerable_package": NamedFieldTextItem(
                        name="Vulnerable Package", value=f"{purl.name}:{purl.version}"
                    ),
                    "found_by": NamedFieldTextItem(
                        name="Found by", value=" | ".join([comp.name for comp in vuln.tools.components])
                    ),
                },
            )
            for purl in purls
        ]

        return gitlab_vulns



"""Setup hoppr-cop as a plugin for hoppr.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/hoppr_plugin/hopprcop_plugin.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Final

from hoppr import BomAccess, HopprPlugin, Result, Vulnerability, hoppr_process

import hopprcop

from hopprcop.combined.combined_scanner import CombinedScanner
from hopprcop.reporting import Reporting
from hopprcop.reporting.models import ReportFormat


if TYPE_CHECKING:
    from hoppr import HopprContext


class HopprCopPlugin(HopprPlugin):
    """Hoppr plugin wrapper for hoppr-cop integration."""

    class ComponentVulnerabilityWrapper:
        """Wrapper for the vulnerabilities associated with a component."""

        def __init__(
            self,
            serial_number: str,
            version: int,
            vulnerabilities: list[Vulnerability] | None = None,
        ):
            self.serial_number = serial_number
            self.version = version
            self.vulnerabilities = vulnerabilities or []

    products: Final[list[str]] = ["generic/*", "generic/hopprcop-vulnerability-results-details/*"]  # type: ignore[misc]

    EMBEDDED_VEX: Final[str] = "embedded_cyclone_dx_vex"
    LINKED_VEX: Final[str] = "linked_cyclone_dx_vex"

    bom_access = BomAccess.FULL_ACCESS

    DEFAULT_SCANNERS: Final[list[str]] = [
        "hopprcop.gemnasium.gemnasium_scanner.GemnasiumScanner",
        "hopprcop.grype.grype_scanner.GrypeScanner",
        "hopprcop.trivy.trivy_scanner.TrivyScanner",
        "hopprcop.ossindex.oss_index_scanner.OSSIndexScanner",
    ]

    def __init__(self, context: HopprContext, config: dict | None = None) -> None:
        super().__init__(context, config)

        self.config = self.config or {}
        frmts: list[str] = self.config.get("result_formats", [self.EMBEDDED_VEX])
        self.formats: list[ReportFormat] = []

        # Correct format strings to bridge legacy naming and cli naming
        if self.EMBEDDED_VEX in frmts:
            self.formats.append(ReportFormat.CYCLONE_DX)
        if self.LINKED_VEX in frmts:
            self.formats.append(ReportFormat.LINKED_VEX)
        self.formats.extend(
            [ReportFormat[fmt.upper()] for fmt in frmts if fmt not in {self.LINKED_VEX, self.EMBEDDED_VEX}]
        )

        output_dir = Path(self.config.get("output_dir", self.context.collect_root_dir / "generic"))
        output_dir.mkdir(parents=True, exist_ok=True)

        self.reporting = Reporting(
            output_path=output_dir,
            base_name=self.config.get("base_report_name", "hopprcop-vulnerability-results"),
        )

        self.results: list[Vulnerability] = []

    def get_version(self) -> str:
        """__version__ required for all HopprPlugin implementations."""
        return hopprcop.__version__

    @hoppr_process
    def pre_stage_process(self) -> Result:
        """Supply SBOM to hoppr cop to perform vulnerabilty check."""
        self.get_logger().info("[ Executing hopprcop vulnerability check ]")

        parsed_bom = self.context.delivered_sbom

        combined_scanner = CombinedScanner()
        combined_scanner.set_scanners((self.config or {}).get("scanners", self.DEFAULT_SCANNERS))

        try:
            self.results = combined_scanner.get_vulnerabilities_for_sbom(parsed_bom)
        except Exception as exc:
            return Result.fail(message=str(exc))

        self.reporting.generate_vulnerability_reports(self.formats, self.results, parsed_bom)

        self.get_logger().flush()

        return Result.success(return_obj=parsed_bom)



"""Defines the common interface for a vulnerability scanner.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/vulnerability_scanner.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import distutils.spawn
import os

from typing import TYPE_CHECKING, ClassVar

import rich

from hoppr import Component, ComponentType, Tools, cdx
from hoppr.utils import get_package_url
from typing_extensions import deprecated


if TYPE_CHECKING:
    from hoppr import Sbom, Vulnerability
    from packageurl import PackageURL


class VulnerabilityMeta(type):
    """Metaclass to define interface expectations of a vulnerability scanner."""

    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        return cls.__subclasscheck__(type(instance))

    @classmethod
    def __subclasscheck__(cls, subclass: type) -> bool:
        return all(
            [
                callable(getattr(subclass, "get_vulnerabilities_by_purl", None)),
                callable(getattr(subclass, "get_vulnerabilities_by_sbom", None)),
                callable(getattr(subclass, "should_activate", None)),
                getattr(subclass, "_scanner_author", None),
                getattr(subclass, "_scanner_name", None),
            ]
        )


class VulnerabilitySuper(metaclass=VulnerabilityMeta):
    """A super class that defines the expected interface for a vulnerability scanner."""

    _scanner_author: str
    _scanner_name: str

    required_tools_on_path: ClassVar[list[str]] = []
    required_environment_variables: ClassVar[list[str]] = []
    offline_mode_supported: bool = False
    offline_mode: bool = False

    def __init_subclass__(
        cls, author: str | None = None, name: str | None = None, offline_mode_supported: bool = False
    ) -> None:
        super().__init_subclass__()

        cls.offline_mode_supported = offline_mode_supported

        # Set class attributes from keyword args if not set in class definition
        for attr_name in {"_scanner_author", "_scanner_name"}:
            attr_value = getattr(cls, attr_name, None) or locals().get(attr_name.removeprefix("_scanner_"))
            setattr(cls, attr_name, attr_value)

            if not getattr(cls, attr_name):
                raise NotImplementedError(f"{cls.__name__} must implement '{attr_name}' class attribute")

    def get_vulnerability_db(self) -> bool:
        """Downloads vulnerability database.

        Return status of downloading the database as a boolean.
        """
        return False

    def get_vulnerabilities_for_purl(self, purls: list[str]) -> list[Vulnerability]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a list of CycloneDX vulnerabilities.
        """
        return []

    def get_vulnerabilities_for_sbom(self, bom: Sbom) -> list[Vulnerability]:
        """Get the vulnerabilities for a CycloneDx compatible Software Bill of Materials (SBOM).

        Returns a list of CycloneDX vulnerabilities.
        """
        purls = [component.purl for component in bom.components or [] if component.purl]

        return self.get_vulnerabilities_for_purl(purls)

    @deprecated(
        "VulnerabilitySuper.get_vulnerabilities_by_purl is deprecated, use VulnerabilitySuper.get_vulnerabilities_for_purl"
    )
    def get_vulnerabilities_by_purl(self, purls: list[PackageURL]) -> dict[str, list[Vulnerability]]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        return {}

    @deprecated(
        "VulnerabilitySuper.get_vulnerabilities_by_sbom is deprecated, use VulnerabilitySuper.get_vulnerabilities_for_sbom"
    )
    def get_vulnerabilities_by_sbom(self, bom: Sbom) -> dict[str, list[Vulnerability]]:
        """Accepts a cyclone dx compatible SBOM and returns a list of vulnerabilities.

        Return a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        purls = [get_package_url(component.purl) for component in bom.components or [] if component.purl]

        return self.get_vulnerabilities_by_purl(purls)

    def scanner_tools(self) -> Tools:
        """Return `Tools` object with a component describing the scanner tool."""
        return Tools(
            components=[
                Component(
                    name=self._scanner_name,
                    supplier=cdx.OrganizationalEntity(name=self._scanner_author),
                    type=ComponentType.APPLICATION,
                )
            ]
        )

    def should_activate(self) -> bool:
        """Checks if the vulnerability scanner should activate based on it's requirements."""
        activate = True

        if missing_tools := [tool for tool in self.required_tools_on_path if not distutils.spawn.find_executable(tool)]:
            activate = False
            rich.print(f"{type(self).__name__} is not activated because required tools are not on the PATH:")
            rich.print(f"{', '.join(missing_tools)}")

        if missing_env := set(self.required_environment_variables).difference(os.environ):
            activate = False
            rich.print(f"{type(self).__name__} is not activated because required environment variables are not set:")
            rich.print(f"{', '.join(missing_env)}")

        if self.offline_mode and not self.offline_mode_supported:
            activate = False
            rich.print(f"{type(self).__name__} is not activated because offline mode is not supported by this scanner")

        return activate



"""A Vulnerability Scanner that combines results from all configured scanners.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/combined/combined_scanner.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import concurrent.futures
import importlib
import warnings

from typing import TYPE_CHECKING, ClassVar, TypeVar, cast

import rich

from hoppr import Sbom, Vulnerability
from typing_extensions import deprecated

from hopprcop.enhancements.analysis_assessment import Analysis
from hopprcop.vulnerability_combiner import combine_vulnerabilities, merge_scanner_vulnerabilities
from hopprcop.vulnerability_scanner import VulnerabilitySuper


if TYPE_CHECKING:
    from collections.abc import Callable

    from packageurl import PackageURL

    from hopprcop.vulnerability_enhancer import BaseEnhancer


ScanResultT = TypeVar("ScanResultT", list[Vulnerability], bool)
ScanResultDeprecatedT = TypeVar("ScanResultDeprecatedT", dict[str, list[Vulnerability]], bool)


# Put here to ensure this is set no matter how it is run
def _showwarning(message: Warning | str, category: type[Warning], *_, **__):
    rich.print(f"[bold yellow]{category.__name__}: {message}")


# Override `warnings.showwarning` to print to console
warnings.showwarning = _showwarning
warnings.filterwarnings(action="once", category=DeprecationWarning)


class ScannerError(RuntimeError):
    """Error class that is thrown when an exception is encountered in a scanner."""


class CombinedScanner(VulnerabilitySuper, author="hoppr-cop", name="CombinedScanner"):
    """A Vulnerability Scanner that combines results from all configured scanners."""

    scanners: ClassVar[list[VulnerabilitySuper]] = []
    enhancers: ClassVar[list[BaseEnhancer]] = []
    assessment = Analysis()

    def set_scanners(self, scanners: list[VulnerabilitySuper] | list[str]):
        """Sets the scanners that should be used for vulnerability scanning.

        The argument can either be a list of scanner instances or a list of fully qualified strings to a scanner
        instance. For example ["vuln.gemnasium.gemnasium_scanner.GemnasiumScanner"].
        """
        for scanner in scanners:
            if isinstance(scanner, str):
                modname, _, clsname = scanner.rpartition(".")
                mod = importlib.import_module(modname)
                scanner = cast("VulnerabilitySuper", getattr(mod, clsname)())

            if scanner.should_activate():
                rich.print(f"{scanner.__class__.__name__} is activated")
                self.scanners.append(scanner)

    def set_enhancers(self, enhancers: list[BaseEnhancer]):
        """Sets the scanners that should be used for vulnerability scanning.

        Args:
            enhancers: Either a list of scanner instances or fully qualified scanner class name strings
        instance. For example ["vuln.gemnasium.gemnasium_scanner.GemnasiumScanner"].
        """
        self.enhancers.extend(enhancer for enhancer in enhancers if enhancer.should_activate())

    def _run_concurrently(self, function: Callable[[VulnerabilitySuper], ScanResultT]) -> dict[str, ScanResultT]:
        results: dict[str, ScanResultT] = {}

        with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
            futures = {executor.submit(function, scanner): scanner for scanner in self.scanners}

        for future in concurrent.futures.as_completed(futures):
            scanner = type(futures[future]).__name__
            try:
                result: ScanResultT = future.result()
                results[scanner] = result
            except Exception as exc:
                rich.print(f"{scanner} generated an exception: {exc}")
                raise ScannerError(exc) from exc

        return results

    @deprecated("CombinedScanner._run_concurrently_deprecated is deprecated, use CombinedScanner._run_concurrently")
    def _run_concurrently_deprecated(
        self, function: Callable[[VulnerabilitySuper], ScanResultDeprecatedT]
    ) -> list[ScanResultDeprecatedT]:
        results: list[ScanResultDeprecatedT] = []

        with concurrent.futures.ThreadPoolExecutor(max_workers=6) as executor:
            futures = {executor.submit(function, scanner): scanner for scanner in self.scanners}

        for future in concurrent.futures.as_completed(futures):
            try:
                result: ScanResultDeprecatedT = future.result()
                if self.assessment.enabled and not isinstance(result, bool):
                    results.append(self.assessment.check_assessments(result, type(future).__name__))
                else:
                    results.append(result)
            except Exception as exc:
                rich.print(f"{type(futures[future]).__name__} generated an exception: {exc}")
                raise exc

        return results

    def _apply_enhancements(self, vuln_list: list[Vulnerability]) -> list[Vulnerability]:
        if self.assessment.enabled:
            self.assessment.apply_assessments(vuln_list)

        for enhancer in self.enhancers:
            enhancer.enhance_vulnerabilities(vuln_list)

        return vuln_list

    def get_vulnerabilities_for_purl(self, purls: list[str]) -> list[Vulnerability]:  # pragma: no cover
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a list of CycloneDX vulnerabilities.
        """

        def submit_to_scanner_purl(scanner: VulnerabilitySuper) -> list[Vulnerability]:
            return scanner.get_vulnerabilities_for_purl(purls)

        vuln_map: dict[str, list[Vulnerability]] = self._run_concurrently(submit_to_scanner_purl)
        results: list[Vulnerability] = merge_scanner_vulnerabilities(vuln_map)
        return self._apply_enhancements(results)

    def get_vulnerabilities_for_sbom(self, bom: Sbom) -> list[Vulnerability]:  # pragma: no cover
        """Get the vulnerabilities for a CycloneDx compatible Software Bill of Materials (SBOM).

        Returns a list of CycloneDX vulnerabilities.
        """

        def submit_to_scanner(scanner: VulnerabilitySuper) -> list[Vulnerability]:
            return scanner.get_vulnerabilities_for_sbom(bom)

        vuln_map: dict[str, list[Vulnerability]] = self._run_concurrently(submit_to_scanner)
        results: list[Vulnerability] = merge_scanner_vulnerabilities(vuln_map)
        return self._apply_enhancements(results)

    @deprecated(
        "CombinedScanner.get_vulnerabilities_by_purl is deprecated, use CombinedScanner.get_vulnerabilities_for_purl"
    )
    def get_vulnerabilities_by_purl(
        self, purls: list[PackageURL]
    ) -> dict[str, list[Vulnerability]]:  # pragma: no cover
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """

        def submit_to_scanner_purl(scanner: VulnerabilitySuper) -> dict[str, list[Vulnerability]]:
            return scanner.get_vulnerabilities_by_purl(purls)

        results: list[dict[str, list[Vulnerability]]] = self._run_concurrently_deprecated(submit_to_scanner_purl)
        return combine_vulnerabilities(list(results))

    @deprecated(
        "CombinedScanner.get_vulnerabilities_by_sbom is deprecated, use CombinedScanner.get_vulnerabilities_for_sbom"
    )
    def get_vulnerabilities_by_sbom(self, bom: Sbom) -> dict[str, list[Vulnerability]]:  # pragma: no cover
        """Parse a CycloneDX compatible SBOM and return a list of vulnerabilities.

        Returns a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """

        def submit_to_scanner(scanner: VulnerabilitySuper) -> dict[str, list[Vulnerability]]:
            return scanner.get_vulnerabilities_by_sbom(bom)

        results = self._run_concurrently_deprecated(submit_to_scanner)
        return combine_vulnerabilities(list(results))

    def get_vulnerability_dbs(self) -> bool:  # pragma: no cover
        """Load the vulnerability Databases for enabled scanners.

        Returns a boolean representation of success.
        """

        def submit_to_scanner(scanner: VulnerabilitySuper) -> bool:
            return scanner.get_vulnerability_db()

        results = self._run_concurrently(submit_to_scanner)
        return list(results.values()).count(False) == 0

    def set_assessment_path(self, path: str):  # pragma: no cover
        """Sets the path to the analysis.assessment.yml file.

        Argument path can be relative or fully qualified path
        For example "./" will look for analysis.assessment.yml in the current directory
        """
        self.assessment._set_assessment_path(path)



"""A vulnerability scanner that locates vulnerabilities in the Sonatype OSS Index.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/ossindex/oss_index_scanner.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import contextlib
import os

from copy import deepcopy
from typing import TYPE_CHECKING, ClassVar, cast

import rich

from cvss import CVSS2, CVSS3, CVSS4
from cvss.exceptions import CVSS2MalformedError, CVSS3MalformedError, CVSS4MalformedError
from hoppr import Affect, Vulnerability, cdx
from hoppr.utils import get_package_url
from packageurl import PackageURL
from requests.auth import HTTPBasicAuth
from typing_extensions import deprecated

from hopprcop.ossindex.api.ossindex import OssIndex
from hopprcop.utils import (
    _add_vulnerability,
    get_advisories_from_urls,
    get_vulnerability_source,
    purl_check,
    unsupported_purl_feedback,
)
from hopprcop.vulnerability_scanner import VulnerabilitySuper


if TYPE_CHECKING:
    from hopprcop.ossindex.api.model import OssIndexComponent, Vulnerability as OssVulnerability


class OSSIndexScanner(VulnerabilitySuper, author="Sonatype", name="OSS-Index", offline_mode_supported=False):
    """A vulnerability scanner that locates vulnerabilities in Sonotypes' OSS Index."""

    required_environment_variables: ClassVar[list[str]] = ["OSS_INDEX_TOKEN", "OSS_INDEX_USER"]

    api = OssIndex()
    api.oss_index_authentication = HTTPBasicAuth(
        username=os.getenv("OSS_INDEX_USER", ""),
        password=os.getenv("OSS_INDEX_TOKEN", ""),
    )

    supported_types: ClassVar[list[str]] = [
        "cargo",
        "conan",
        "gem",
        "golang",
        "maven",
        "npm",
        "nuget",
        "pypi",
        "rpm",
    ]

    def __init__(self, offline_mode: bool = False):
        self.offline_mode = offline_mode
        super().__init__()

    def get_vulnerabilities_for_purl(self, purls: list[str]) -> list[Vulnerability]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a list of CycloneDX vulnerabilities.
        """
        # Check for unsupported purl types and report if found
        filtered_purls = list(
            filter(lambda x: x.type in self.supported_types, [get_package_url(purl) for purl in purls])
        )
        unsupported_purl_feedback(self._scanner_name, self.supported_types, filtered_purls)
        packages_with_version = [package for package in filtered_purls if package.version is not None]
        if packages_without_version := [package for package in filtered_purls if package.version is None]:
            rich.print(
                "[yellow]WARNING: OSS-Index -- does not support purls with missing versions, components may be missed."
            )
            for package in packages_without_version:
                rich.print(f"[yellow]  {package}")

        if not packages_with_version:
            return []

        cleaned_purl_map: dict[str, str] = {}

        def remove_qualifiers(pkg_url: PackageURL) -> PackageURL:
            before_cleaning = pkg_url.to_string()
            cleaned_purl = deepcopy(pkg_url)
            cast("dict", cleaned_purl.qualifiers).clear()

            if cleaned_purl.type == "rpm":
                cleaned_purl = PackageURL(name=pkg_url.name, type=pkg_url.type, version=pkg_url.version)

            cleaned_purl_map[cleaned_purl.to_string()] = before_cleaning

            return cleaned_purl

        cleaned_purls = list(map(remove_qualifiers, packages_with_version))

        matches: list[OssIndexComponent] = self.api.get_component_report(packages=cleaned_purls)
        results: dict[str, Vulnerability] = {}

        for match in matches:
            for vulnerability in match.vulnerabilities:
                _add_vulnerability(
                    vulnerability.id,
                    results,
                    self._convert_to_cyclone_dx(vulnerability, cleaned_purl_map[match.coordinates], purls),
                )

        return list(results.values())

    @deprecated(
        "OSSIndexScanner.get_vulnerabilities_by_purl is deprecated, use OSSIndexScanner.get_vulnerabilities_for_purl"
    )
    def get_vulnerabilities_by_purl(self, purls: list[PackageURL]) -> dict[str, list[Vulnerability]]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        # Check for unsupported purl types and report if found
        unsupported_purl_feedback(self._scanner_name, self.supported_types, purls)
        purls = list(filter(lambda x: x.type in self.supported_types, purls))
        cleaned_purl_map: dict[str, str] = {}

        def remove_qualifiers(pkg_url: PackageURL) -> PackageURL:
            before_cleaning = pkg_url.to_string()
            cast("dict", pkg_url.qualifiers).clear()

            if pkg_url.type == "rpm":
                pkg_url = PackageURL(name=pkg_url.name, type=pkg_url.type, version=pkg_url.version)

            cleaned_purl_map[pkg_url.to_string()] = before_cleaning

            return pkg_url

        purls = list(map(remove_qualifiers, purls))

        results: list[OssIndexComponent] = self.api.get_component_report(packages=purls)
        enhanced_results: dict[str, list[Vulnerability]] = {}

        for result in results:
            purl = result.coordinates
            enhanced_results[cleaned_purl_map[purl]] = []

            for vulnerability in result.vulnerabilities:
                enhanced_results[cleaned_purl_map[purl]].append(self._convert_to_cyclone_dx_deprecated(vulnerability))

        return enhanced_results

    def _convert_to_cyclone_dx(
        self, vulnerability: OssVulnerability, vuln_purl: str, original_purls: list[str]
    ) -> Vulnerability:
        """Convert an OSS Index vulnerability to cyclone dx."""
        vuln_id = vulnerability.cve if vulnerability.cve is not None else vulnerability.display_name

        try:
            cwes = [int((vulnerability.cwe or "").removeprefix("CWE-"))]
        except ValueError:
            cwes = []

        cyclone_vuln = Vulnerability(
            id=vuln_id,
            description=vulnerability.description,
            cwes=cwes,
            source=get_vulnerability_source(vuln_id),
        )

        cyclone_vuln.ratings = []

        # Add Affects for the current purl to the vulnerability
        cyclone_vuln.affects = [
            Affect.parse_obj(
                {
                    "ref": purl,
                    "versions": [{"version": get_package_url(purl).version, "status": "affected"}],
                }
            )
            for purl in original_purls
            if purl_check(vuln_purl, purl)
        ]

        if vulnerability.cvss_vector is not None:
            cvss = None
            cvss_types_iterator = iter([CVSS4, CVSS3, CVSS2])
            try:
                while cvss is None:
                    cvss_type = next(cvss_types_iterator)
                    with contextlib.suppress(CVSS4MalformedError, CVSS3MalformedError, CVSS2MalformedError):
                        cvss = cvss_type(vulnerability.cvss_vector)
                score_method = {
                    CVSS4: cdx.ScoreMethod.CVSSv4,
                    CVSS3: cdx.ScoreMethod.CVSSv3,
                    CVSS2: cdx.ScoreMethod.CVSSv2,
                }[type(cvss)]
                cyclone_vuln.ratings.append(
                    cdx.Rating(
                        score=float(cvss.base_score or 0.0),
                        severity=cdx.Severity[cvss.severities()[0].lower()],
                        method=score_method,
                        vector=cvss.vector,
                    )
                )
            except StopIteration:
                rich.print(
                    f"[yellow]Vulnerability {cyclone_vuln.id} contains "
                    f"unrecognized CVSS vector {vulnerability.cvss_vector}"
                )

        cyclone_vuln.advisories = get_advisories_from_urls(list(vulnerability.external_references))
        cyclone_vuln.tools = self.scanner_tools()

        return cyclone_vuln

    @deprecated(
        "OSSIndexScanner._convert_to_cyclone_dx_deprecated is deprecated, use OSSIndexScanner._convert_to_cyclone_dx"
    )
    def _convert_to_cyclone_dx_deprecated(self, vulnerability: OssVulnerability) -> Vulnerability:
        """Convert an OSS Index vulnerability to cyclone dx."""
        vuln_id = vulnerability.cve if vulnerability.cve is not None else vulnerability.display_name

        try:
            cwes = [int((vulnerability.cwe or "").removeprefix("CWE-"))]
        except ValueError:
            cwes = []

        cyclone_vuln = Vulnerability(
            id=vuln_id,
            description=vulnerability.description,
            cwes=cwes,
            source=get_vulnerability_source(vuln_id),
        )

        cyclone_vuln.ratings = []

        if vulnerability.cvss_vector is not None:
            try:
                cvss = CVSS3(vulnerability.cvss_vector)
                cyclone_vuln.ratings.append(
                    cdx.Rating(
                        score=cvss.base_score,
                        severity=cdx.Severity[cvss.severities()[0].lower()],
                        method=cdx.ScoreMethod.CVSSv3,
                        vector=cvss.vector,
                    )
                )
            except CVSS3MalformedError:
                cvss = CVSS2(vulnerability.cvss_vector)
                cyclone_vuln.ratings.append(
                    cdx.Rating(
                        score=cvss.base_score,
                        severity=cdx.Severity[cvss.severities()[0].lower()],
                        method=cdx.ScoreMethod.CVSSv2,
                        vector=cvss.vector,
                    )
                )

        cyclone_vuln.advisories = get_advisories_from_urls(list(vulnerability.external_references))
        cyclone_vuln.tools = self.scanner_tools()

        return cyclone_vuln



"""This file contains auto generated pydantic models for the grype json output.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/grype/models.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

from typing import Annotated, Any

from pydantic import BaseModel, Field


class GrypeBaseModel(BaseModel, allow_population_by_field_name=True): ...


class Metrics(GrypeBaseModel):
    base_score: Annotated[float | None, Field(alias="baseScore")] = None
    exploitability_score: Annotated[float | None, Field(alias="exploitabilityScore")] = None
    impact_score: Annotated[float | None, Field(alias="impactScore")] = None


class VendorMetadata(GrypeBaseModel):
    base_severity: str | None = None
    status: str | None = None


class Cvs(GrypeBaseModel):
    source: str | None = None
    type: str | None = None
    version: str
    vector: str
    metrics: Metrics | None = None
    vendor_metadata: Annotated[VendorMetadata | None, Field(alias="vendorMetadata")] = None


class Fix(GrypeBaseModel):
    versions: list[str] = []
    state: str = ""


class Vulnerability(GrypeBaseModel):
    id: str
    data_source: Annotated[str | None, Field(alias="dataSource")] = None
    namespace: str | None = None
    urls: list[str] = []
    description: str | None = None
    severity: str | None = None
    cvss: list[Cvs] = []
    fix: Fix = Fix()
    advisories: list | None = None


class Package(GrypeBaseModel):
    name: str | None = None
    version: str | None = None


class SearchedBy(GrypeBaseModel):
    language: str | None = None
    namespace: str | None = None
    package: Package | None = None


class Found(GrypeBaseModel):
    version_constraint: Annotated[str | None, Field(alias="versionConstraint")] = None
    vulnerability_id: Annotated[str | None, Field(alias="vulnerabilityID")] = None


class MatchDetail(GrypeBaseModel):
    type: str | None = None
    matcher: str | None = None
    searched_by: Annotated[SearchedBy | None, Field(alias="searchedBy")] = None
    found: Found | None = None


class Artifact(GrypeBaseModel):
    id: str | None = None
    name: str | None = None
    version: str | None = None
    type: str | None = None
    locations: list | None = None
    language: str | None = None
    licenses: list | None = None
    cpes: list | None = None
    purl: str
    upstreams: list | None = None


class Match(GrypeBaseModel):
    vulnerability: Vulnerability
    related_vulnerabilities: Annotated[list[Vulnerability], Field(alias="relatedVulnerabilities")]
    match_details: Annotated[list[MatchDetail] | None, Field(alias="matchDetails")] = None
    artifact: Artifact


class Source(GrypeBaseModel):
    type: str | None = None
    target: str | dict[str, Any] | None = None


class Distro(GrypeBaseModel):
    name: str | None = None
    version: str | None = None
    id_like: Annotated[list[str] | None, Field(alias="idLike")] = None


class Search(GrypeBaseModel):
    scope: str | None = None
    unindexed_archives: Annotated[bool | None, Field(alias="unindexed-archives")] = None
    indexed_archives: Annotated[bool | None, Field(alias="indexed-archives")] = None


class DatabaseConfig(GrypeBaseModel):
    cache_dir: Annotated[str | None, Field(alias="cache-dir")] = None
    update_url: Annotated[str | None, Field(alias="update-url")] = None
    ca_cert: Annotated[str | None, Field(alias="ca-cert")] = None
    auto_update: Annotated[bool | None, Field(alias="auto-update")] = None
    validate_by_hash_on_start: Annotated[bool | None, Field(alias="validate-by-hash-on-start")] = None
    validate_age: Annotated[bool | None, Field(alias="validate-age")] = None
    max_allowed_built_age: Annotated[int | None, Field(alias="max-allowed-built-age")] = None


class Maven(GrypeBaseModel):
    search_upstream_by_sha1: Annotated[bool | None, Field(alias="searchUpstreamBySha1")] = None
    base_url: Annotated[str | None, Field(alias="baseUrl")] = None


class ExternalSources(GrypeBaseModel):
    enable: bool | None = None
    maven: Maven | None = None


class LanguageConfigMatch(GrypeBaseModel):
    using_cpes: Annotated[bool | None, Field(alias="using-cpes")] = None
    always_use_cpe_for_stdlib: Annotated[bool | None, Field(alias="always-use-cpe-for-stdlib")] = None


class ConfigMatch(GrypeBaseModel):
    java: LanguageConfigMatch | None = None
    dotnet: LanguageConfigMatch | None = None
    golang: LanguageConfigMatch | None = None
    javascript: LanguageConfigMatch | None = None
    python: LanguageConfigMatch | None = None
    ruby: LanguageConfigMatch | None = None
    rust: LanguageConfigMatch | None = None
    stock: LanguageConfigMatch | None = None


class Dev(GrypeBaseModel):
    profile_cpu: Annotated[bool | None, Field(alias="profile-cpu")] = None
    profile_mem: Annotated[bool | None, Field(alias="profile-mem")] = None


class Registry(GrypeBaseModel):
    insecure_skip_tls_verify: Annotated[bool | None, Field(alias="insecure-skip-tls-verify")] = None
    insecure_use_http: Annotated[bool | None, Field(alias="insecure-use-http")] = None
    auth: list | None = None


class Log(GrypeBaseModel):
    structured: bool | None = None
    level: str | None = None
    file: str | None = None


class Configuration(GrypeBaseModel):
    config_path: Annotated[str | None, Field(alias="configPath")] = None
    verbosity: int | None = None
    output: list[str] = []
    file: str | None = None
    distro: str | None = None
    add_cpes_if_none: Annotated[bool | None, Field(alias="add-cpes-if-none")] = None
    output_template_file: Annotated[str | None, Field(alias="output-template-file")] = None
    quiet: bool | None = None
    check_for_app_update: Annotated[bool | None, Field(alias="check-for-app-update")] = None
    only_fixed: Annotated[bool | None, Field(alias="only-fixed")] = None
    only_notfixed: Annotated[bool | None, Field(alias="only-notfixed")] = None
    platform: str | None = None
    search: Search | None = None
    ignore: list | None = None
    exclude: list | None = None
    db: DatabaseConfig | None = None
    external_sources: Annotated[ExternalSources | None, Field(alias="externalSources")] = None
    match: ConfigMatch | None = None
    dev: Dev | None = None
    fail_on_severity: Annotated[str | None, Field(alias="fail-on-severity")] = None
    registry: Registry | None = None
    log: Log | None = None
    show_suppressed: Annotated[bool | None, Field(alias="show-suppressed")] = None
    by_cve: Annotated[bool | None, Field(alias="by-cve")] = None
    name: str | None = None
    default_image_pull_source: Annotated[str | None, Field(alias="default-image-pull-source")] = None


class Database(GrypeBaseModel):
    built: str | None = None
    schema_version: Annotated[int | None, Field(alias="schemaVersion")] = None
    location: str | None = None
    checksum: str | None = None
    error: None


class Descriptor(GrypeBaseModel):
    name: str | None = None
    version: str | None = None
    configuration: Configuration | None = None
    db: Database | None = None
    timestamp: str | None = None


class GrypeResult(GrypeBaseModel):
    matches: list[Match]
    source: Source | None = None
    distro: Distro | None = None
    descriptor: Descriptor | None = None



"""Scan an SBOM using the Grype CLI.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/grype/grype_scanner.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import json
import os

from subprocess import PIPE, Popen
from typing import TYPE_CHECKING, Any, ClassVar

import rich

from cvss import CVSS2, CVSS3
from hoppr import Affect, HopprError, Sbom, Vulnerability, cdx
from hoppr.utils import get_package_url
from typing_extensions import deprecated

from hopprcop.grype.models import GrypeResult, Match, Vulnerability as GrypeVulnerability
from hopprcop.utils import (
    _add_vulnerability,
    build_bom_from_purls,
    create_bom_from_purl_list,
    get_advisories_from_urls,
    get_references_from_ids,
    get_vulnerability_source,
    purl_check,
    unsupported_purl_feedback,
)
from hopprcop.vulnerability_scanner import VulnerabilitySuper


if TYPE_CHECKING:
    from packageurl import PackageURL


class GrypeScanner(VulnerabilitySuper, author="Anchore", name="Grype", offline_mode_supported=True):
    """This scanner utilizes the anchore grype command line to gather vulnerabilities."""

    required_tools_on_path: ClassVar[list[str]] = ["grype"]
    grype_os_distro = os.getenv("OS_DISTRIBUTION", None)

    process_environment = None
    supported_types: ClassVar[list[str]] = [
        "cargo",
        "composer",
        "docker",
        "gem",
        "golang",
        "maven",
        "npm",
        "nuget",
        "oci",
        "pypi",
        "rpm",
    ]

    def __init__(self, offline_mode: bool = False, grype_os_distro: str | None = None):
        self.grype_os_distro = grype_os_distro or os.getenv("OS_DISTRIBUTION", None)

        self.offline_mode = offline_mode
        self.process_environment = os.environ.copy()
        if self.offline_mode:
            self.process_environment["GRYPE_DB_AUTO_UPDATE"] = "false"
            self.process_environment["GRYPE_DB_VALIDATE_AGE"] = "false"

        super().__init__()

    def get_vulnerability_db(self) -> bool:
        """Downloads vulnerability database."""
        args = ["grype", "db", "update"]

        with Popen(args, stdout=PIPE, stderr=PIPE) as process:
            stdout, stderr = process.communicate()

            if not stdout and stderr:
                rich.print(f"GrypeScanner: generated an exception: {stderr.decode()}")
                return False

            rich.print(f"GrypeScanner: {stdout.decode()}")
            return True

    def get_vulnerabilities_for_purl(self, purls: list[str]) -> list[Vulnerability]:  # pragma: no cover
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a list of CycloneDX vulnerabilities.
        """
        bom = create_bom_from_purl_list(purls)
        return self.get_vulnerabilities_for_sbom(bom)

    def get_vulnerabilities_for_sbom(self, bom: Sbom) -> list[Vulnerability]:
        """Get the vulnerabilities for a CycloneDx compatible Software Bill of Materials (SBOM).

        Returns a list of CycloneDX vulnerabilities.
        """
        # Check for unsupported purl types and report if found
        purls = [component.purl for component in bom.components or [] if component.purl]
        unsupported_purl_feedback(self._scanner_name, self.supported_types, [get_package_url(purl) for purl in purls])

        results: dict[str, Vulnerability] = {}

        args = ["grype", "--output", "json"]
        if self.grype_os_distro is not None:
            args += ["--distro", self.grype_os_distro]

        with Popen(args, stdout=PIPE, stdin=PIPE, stderr=PIPE, env=self.process_environment) as process:
            # Remove tools from metadata due to cyclonedx-go only having partial support for spec version 1.5 (as of 0.72.0)
            # TODO: roll back once full support is in cyclonedx-go
            parsed_bom = bom.copy(deep=True)
            if parsed_bom.metadata:
                parsed_bom.metadata.tools = None

            stdout, stderr = process.communicate(input=(bytes(parsed_bom.json(), "utf-8")))
            if process.returncode != 0:
                raise HopprError(f"{self.__class__.__name__} generated an exception: {stderr.decode()}")

            result = GrypeResult(**json.loads(stdout))

            for match in [match for match in result.matches if match.artifact.purl in purls]:
                match_purl = get_package_url(match.artifact.purl)
                if match_purl.type != "npm" or match_purl.namespace != "@types":
                    try:
                        _add_vulnerability(match.vulnerability.id, results, self._convert_to_cyclone_dx(match, purls))
                    except AttributeError:
                        rich.print(
                            f"[yellow]WARNING: {self._scanner_name} -- received an AttributeError while attempting to add vulnerability - {match.vulnerability.id}"
                        )

        return list(results.values())

    @deprecated("GrypeScanner.get_vulnerabilities_by_purl is deprecated, use GrypeScanner.get_vulnerabilities_for_purl")
    def get_vulnerabilities_by_purl(
        self, purls: list[PackageURL]
    ) -> dict[str, list[Vulnerability]]:  # pragma: no cover
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        bom = build_bom_from_purls(purls)
        return self.get_vulnerabilities_by_sbom(bom)

    @deprecated("GrypeScanner.get_vulnerabilities_by_sbom is deprecated, use GrypeScanner.get_vulnerabilities_for_sbom")
    def get_vulnerabilities_by_sbom(self, bom: Sbom) -> dict[str, list[Vulnerability]]:
        """Parse a CycloneDX compatible SBOM and return a list of vulnerabilities.

        Returns a dictionary of package URL to vulnerabilities.
        """
        args = ["grype", "--output", "json"]
        if self.grype_os_distro is not None:
            args += ["--distro", self.grype_os_distro]

        with Popen(args, stdout=PIPE, stdin=PIPE, stderr=PIPE, env=self.process_environment) as process:
            # Remove tools from metadata due to cyclonedx-go only having partial support for spec version 1.5 (as of 0.72.0)
            # TODO: roll back once full support is in cyclonedx-go
            parsed_bom = bom.copy(deep=True)
            if parsed_bom.metadata:
                parsed_bom.metadata.tools = None

            stdout, stderr = process.communicate(input=(bytes(parsed_bom.json(), "utf-8")))
            result = GrypeResult(**json.loads(stdout))

            results: dict[str, list[Vulnerability]] = {}

        # Use a generator to get all of the component purls if they exist and intialize the results for the purl
        purls = [component.purl for component in bom.components or [] if component.purl]
        for pkg_url in purls:
            results[pkg_url] = []

        # Check for unsupported purl types and report if found
        unsupported_purl_feedback(self._scanner_name, self.supported_types, [get_package_url(purl) for purl in purls])

        for match in list(result.matches):
            purl = get_package_url(match.artifact.purl)

            if purl.type != "npm" or purl.namespace != "@types":
                # Raise the error if `match.artifact.purl` is not already a key in the dict
                # This can occur when the `match.artifact.purl` is not in the `component.purl` above
                try:
                    # Append the Grype result information to the results dict at the purl key
                    results[match.artifact.purl].append(self._convert_to_cyclone_dx_deprecated(match))
                except KeyError:
                    rich.print(f"[yellow]WARNING: {self._scanner_name} -- Match not found: {match.artifact.purl}")

        return results

    def _convert_to_cyclone_dx(self, match: Match, original_purls: list[str]) -> Vulnerability:
        """Converts a match to a vulnerability."""
        related: GrypeVulnerability = next(
            (related_vuln for related_vuln in match.related_vulnerabilities if related_vuln.id.startswith("CVE")),
            (match.related_vulnerabilities[0] if match.related_vulnerabilities else match.vulnerability),
        )

        cyclone_vuln = Vulnerability(
            description=related.description,
            recommendation=(
                f"State: {match.vulnerability.fix.state} | Fix Versions: {','.join(match.vulnerability.fix.versions)}"
            ),
            source=get_vulnerability_source(related.id),
        )

        cyclone_vuln.id = related.id
        cyclone_vuln.ratings = []

        ids = [match.vulnerability.id, *[x.id for x in match.related_vulnerabilities]]

        # Maintain cyclone_vul.source or initialize to empty VulnerabilitySource
        cyclone_vuln.source = cyclone_vuln.source or cdx.VulnerabilitySource()
        cyclone_vuln.source.url = related.data_source
        cyclone_vuln.advisories = get_advisories_from_urls(related.urls)
        cyclone_vuln.references = get_references_from_ids(ids, cyclone_vuln.id)

        affected_purls = [purl for purl in original_purls if purl_check(match.artifact.purl, purl)]
        for purl in affected_purls:
            affect_dict: dict[str, Any] = {"ref": purl}
            if match.artifact.version is not None:
                affect_dict["versions"] = [{"version": match.artifact.version, "status": "affected"}]

            cyclone_vuln.affects.append(Affect.parse_obj(affect_dict))

        cvss_scores = match.vulnerability.cvss or related.cvss

        for cvss in cvss_scores:
            if cvss.version.startswith("3"):
                cvss3 = CVSS3(cvss.vector)
                method = "CVSSv31" if cvss.version == "3.1" else "CVSSv3"

                cyclone_vuln.ratings.append(
                    cdx.Rating(
                        score=float(cvss3.base_score or 0.0),
                        severity=cdx.Severity[cvss3.severities()[0].lower()],
                        method=cdx.ScoreMethod(method),
                        vector=cvss.vector,
                    )
                )
            elif cvss.version.startswith("2"):
                cvss2 = CVSS2(cvss.vector)

                cyclone_vuln.ratings.append(
                    cdx.Rating(
                        score=float(cvss2.base_score or 0.0),
                        severity=cdx.Severity[cvss2.severities()[0].lower()],
                        method=cdx.ScoreMethod.CVSSv2,
                        vector=cvss.vector,
                    )
                )

        if not cyclone_vuln.ratings and match.vulnerability.severity:
            cyclone_vuln.ratings.append(
                cdx.Rating(
                    severity=getattr(cdx.Severity, match.vulnerability.severity.lower(), cdx.Severity.INFO),
                    method=cdx.ScoreMethod.OTHER,
                )
            )

        cyclone_vuln.tools = self.scanner_tools()

        return cyclone_vuln

    @deprecated("GrypeScanner._convert_to_cyclone_dx_deprecated is deprecated, use GrypeScanner._convert_to_cyclone_dx")
    def _convert_to_cyclone_dx_deprecated(self, match: Match) -> Vulnerability:
        """Converts a match to a vulnerability."""
        related: GrypeVulnerability = next(
            (related_vuln for related_vuln in match.related_vulnerabilities if related_vuln.id.startswith("CVE")),
            (match.related_vulnerabilities[0] if match.related_vulnerabilities else match.vulnerability),
        )

        cyclone_vuln = Vulnerability(
            description=related.description,
            recommendation=(
                f"State: {match.vulnerability.fix.state} | Fix Versions: {','.join(match.vulnerability.fix.versions)}"
            ),
            source=get_vulnerability_source(related.id),
        )

        cyclone_vuln.id = related.id
        cyclone_vuln.ratings = []

        ids = [match.vulnerability.id, *[x.id for x in match.related_vulnerabilities]]

        # Maintain cyclone_vul.source or initialize to empty VulnerabilitySource
        cyclone_vuln.source = cyclone_vuln.source or cdx.VulnerabilitySource()
        cyclone_vuln.source.url = related.data_source
        cyclone_vuln.advisories = get_advisories_from_urls(related.urls)
        cyclone_vuln.references = get_references_from_ids(ids, cyclone_vuln.id)
        cvss_scores = match.vulnerability.cvss or related.cvss

        for cvss in cvss_scores:
            if cvss.version.startswith("3"):
                cvss3 = CVSS3(cvss.vector)
                method = "CVSSv31" if cvss.version == "3.1" else "CVSSv3"

                cyclone_vuln.ratings.append(
                    cdx.Rating(
                        score=cvss3.base_score,
                        severity=cdx.Severity[cvss3.severities()[0].lower()],
                        method=cdx.ScoreMethod(method),
                        vector=cvss.vector,
                    )
                )
            elif cvss.version.startswith("2"):
                cvss2 = CVSS2(cvss.vector)

                cyclone_vuln.ratings.append(
                    cdx.Rating(
                        score=cvss2.base_score,
                        severity=cdx.Severity[cvss2.severities()[0].lower()],
                        method=cdx.ScoreMethod.CVSSv2,
                        vector=cvss.vector,
                    )
                )

        if not cyclone_vuln.ratings and match.vulnerability.severity:
            cyclone_vuln.ratings.append(
                cdx.Rating(
                    severity=cdx.Severity[match.vulnerability.severity.lower()],
                    method=cdx.ScoreMethod.OTHER,
                )
            )

        cyclone_vuln.tools = self.scanner_tools()

        return cyclone_vuln



"""Defines the common interface for a vulnerability scanner.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2024 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/vulnerability_enhancer.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import rich


if TYPE_CHECKING:
    from hoppr import Vulnerability


class BaseEnhancer:
    """A super class that defines the expected interface for an enhancer."""

    offline_mode_supported: bool = False
    offline_mode: bool = False
    enabled: bool = False

    def __init_subclass__(cls, offline_mode_supported: bool = False) -> None:
        super().__init_subclass__()

        cls.offline_mode_supported = offline_mode_supported

    def enhance_vulnerabilities(self, vuln_list: list[Vulnerability]):
        """Get the enhancers.

        Return a list of enhancers for vulnerabilities or none if no vulnerabilities are found.
        """

    def should_activate(self) -> bool:
        """Checks if the enhancer should activate based on it's requirements."""
        activate = self.enabled

        if self.enabled and (self.offline_mode and not self.offline_mode_supported):
            activate = False
            rich.print(f"{type(self).__name__} is not activated because offline mode is not supported by this enhancer")

        return activate



"""A Vulnerability Scanner for Gitlab's Gemnasiumm Database.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2022 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/gemnasium/gemnasium_scanner.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import os
import pkgutil
import shutil
import stat
import subprocess
import sys
import tarfile
import tempfile
import zipfile

from datetime import datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar
from urllib.parse import urlparse

import requests
import rich
import yaml

from cvss import CVSS2, CVSS3
from hoppr import Affect, Vulnerability, cdx
from hoppr.utils import get_package_url
from typing_extensions import deprecated

from hopprcop.gemnasium.models import GemnasiumVulnerability
from hopprcop.utils import (
    _add_vulnerability,
    get_advisories_from_urls,
    get_references_from_ids,
    get_vulnerability_source,
    unsupported_purl_feedback,
)
from hopprcop.vulnerability_scanner import VulnerabilitySuper


if TYPE_CHECKING:
    from packageurl import PackageURL


class GemnasiumScanner(VulnerabilitySuper, author="Gitlab", name="Gemnasium", offline_mode_supported=True):
    """A Vulnerability Scanner for Gitlab's Gemnasiumm Database."""

    supported_types: ClassVar[list[str]] = ["conan", "gem", "golang", "gradle", "maven", "npm", "nuget", "pypi"]

    url = os.getenv(
        "GEMNASIUM_DATABASE_ZIP",
        "https://gitlab.com/gitlab-org/advisories-community/-/archive/main/advisories-community-main.tar.gz",
    )

    semver_path = Path("/usr/local/bin/semver")
    required_tools_on_path: ClassVar[list[str]] = ["ruby"]

    def __init__(self, offline_mode: bool = False):
        self.database_path = self._get_cache_dir() / "gemnasium"
        self.offline_mode = offline_mode

        super().__init__()

    def should_activate(self) -> bool:
        """Checks whether the scanner can be activated."""
        activate = super().should_activate()

        if activate:
            if not Path(self.semver_path).exists():
                self._extract_semver_to_local()
            activate = self._download_and_extract_database()

        if not activate and self.offline_mode:
            rich.print(f"{type(self).__name__} is not activated because the offline database doesn't exist on disk")

        return activate

    def get_vulnerability_db(self) -> bool:  # pragma: no cover
        """Downloads vulnerability database."""
        return self._download_and_extract_database()

    def _extract_semver_to_local(self):
        """If the ruby semver command isn't installed then extract from this package."""
        data = (pkgutil.get_data(__name__, "semver") or b"").decode("utf-8")
        self.semver_path = Path(tempfile.gettempdir()) / "semver"
        self.semver_path.write_text(data, encoding="utf-8")
        self.semver_path.chmod(mode=stat.S_IRWXU | stat.S_IRGRP | stat.S_IXGRP | stat.S_IROTH | stat.S_IXOTH)

    def _do_download_and_unpack(self, path_to_compressed_file: Path):
        rich.print(f"Updating Gemnasium database to {self.database_path}")

        request = requests.get(self.url, allow_redirects=True)
        path_to_compressed_file.write_bytes(request.content)

        match path_to_compressed_file.suffix:
            case ".zip":
                with zipfile.ZipFile(path_to_compressed_file, "r") as zip_ref:
                    zip_ref.extractall(self.database_path)
            case tar if tar in {".gz", ".tgz", ".bz2", ".tbz"}:
                with tarfile.open(path_to_compressed_file, "r") as tar_ref:
                    tar_ref.extractall(self.database_path)
            case _:
                rich.print(f"Failed to extract database file: {self.url}")

    def _older_than(self, delta: timedelta, file: Path) -> bool:
        file_time = file.stat().st_mtime
        return datetime.now() - delta > datetime.fromtimestamp(file_time)

    def _download_and_extract_database(self) -> bool:
        """Downloads the gymnasium database."""
        url = urlparse(self.url)

        path_to_compressed_file = self._get_cache_dir() / Path(url.path).name
        file_exists = path_to_compressed_file.exists()
        dest_dir_name = path_to_compressed_file.stem.split(".")[0]

        if not file_exists:
            if self.offline_mode:
                rich.print(
                    f"GemnasiumScanner: Unable to find local database file {path_to_compressed_file.absolute()} for offline execution"
                )
                return False
            self._do_download_and_unpack(path_to_compressed_file)
        elif file_exists and self._older_than(timedelta(hours=24), path_to_compressed_file):
            if not self.offline_mode:
                shutil.rmtree(self.database_path / dest_dir_name)
                self._do_download_and_unpack(path_to_compressed_file)
            else:
                rich.print("GemnasiumScanner: Database file is older than one day for offline mode")
        else:
            rich.print("GemnasiumScanner: Database file is up to date")

        self.database_path = self.database_path / dest_dir_name

        return True

    def _is_affected_range(self, repository_format: str, version: str, affected_range: str) -> bool:
        """Checks if the version matches the affected range based on package manager specific semantic versioning.

        :param repository_format:
        :param version:
        :param affected_range:
        :return:
        """
        try:
            # Gemnasium docs list nuget as supported, however no results are returned
            # nuget package-versioning is based on maven dependency version
            # specification. Below repository_format override allows nuget vuln discovery
            # https://learn.microsoft.com/en-us/nuget/concepts/package-versioning
            # https://maven.apache.org/pom.html#dependency-version-requirement-specification
            repository_format = repository_format.replace("nuget", "maven")

            _win32 = sys.platform == "win32"
            output = subprocess.run(
                [
                    *(["ruby"] if _win32 else []),
                    self.semver_path,
                    "check_version",
                    repository_format,
                    version,
                    affected_range,
                ],
                capture_output=True,
                text=True,
                check=False,
                # TODO this command suddenly started throwing errors, not sure what changed but it needs investigated.
                # It looks like  a spell checker package changed
            )
            return "matches" in str(output)
        except Exception as err:
            rich.print(f"Failed to check version for: {repository_format} {version} {err}")
            return False

    def get_vulnerabilities_for_purl(self, purls: list[str]) -> list[Vulnerability]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a list of CycloneDX vulnerabilities.
        """
        # Check for unsupported purl types and report if found
        unsupported_purl_feedback(self._scanner_name, self.supported_types, [get_package_url(purl) for purl in purls])

        results: dict[str, Vulnerability] = {}

        for purl_str in purls:
            purl = get_package_url(purl_str)
            if not (path := self._get_path(purl)).exists():
                continue

            vuln_files = [file for file in path.glob("*") if file.is_file()]

            for vuln_file in vuln_files:
                try:
                    data = yaml.full_load(vuln_file.read_text(encoding="utf-8"))
                    vuln = GemnasiumVulnerability(**data)

                    if self._is_affected_range(purl.type, purl.version or "", vuln.affected_range):
                        vulnerability = self._convert_to_cyclone_dx(vuln)

                        if vulnerability.ratings:
                            # Add Affects for the current purl to the vulnerability
                            vulnerability.affects = [
                                Affect.parse_obj(
                                    {
                                        "ref": purl_str,
                                        "versions": [{"version": purl.version, "status": "affected"}],
                                    }
                                )
                            ]

                            _add_vulnerability(vulnerability.id or vuln.identifiers[0], results, vulnerability)
                except Exception:
                    rich.print(f"failed to parse gemnasium file for {purl}")
        return list(results.values())

    @deprecated(
        "GemnasiumScanner.get_vulnerabilities_by_purl is deprecated, use GemnasiumScanner.get_vulnerabilities_for_purl"
    )
    def get_vulnerabilities_by_purl(self, purls: list[PackageURL]) -> dict[str, list[Vulnerability]]:
        """Get the vulnerabilities for a list of package URLS (purls).

        Returns a dictionary of package URL to vulnerabilities or none if no vulnerabilities are found.
        """
        # Check for unsupported purl types and report if found
        unsupported_purl_feedback(self._scanner_name, self.supported_types, purls)

        vulnerabilities_by_purl: dict[str, list[Vulnerability]] = {}
        for purl in purls:
            purl_str = purl.to_string()
            vulnerabilities_by_purl[purl_str] = []

            if not (path := self._get_path(purl)).exists():
                continue

            vuln_files = [file for file in path.glob("*") if file.is_file()]

            for vuln_file in vuln_files:
                try:
                    data = yaml.full_load(vuln_file.read_text(encoding="utf-8"))
                    vuln = GemnasiumVulnerability(**data)

                    if self._is_affected_range(purl.type, purl.version or "", vuln.affected_range):
                        vulnerability = self._convert_to_cyclone_dx(vuln)

                        if vulnerability.ratings:
                            vulnerabilities_by_purl[purl_str].append(vulnerability)
                except Exception:
                    rich.print(f"failed to parse gemnasium file for {purl}")

        return vulnerabilities_by_purl

    def _convert_to_cyclone_dx(self, vuln: GemnasiumVulnerability) -> Vulnerability:
        """Converts a gemnasium vulnerabity to a vulnerability."""
        res = list(filter(lambda x: "cve-" in x.lower(), vuln.identifiers))
        vuln_id = res[0] if len(res) > 1 else vuln.identifiers[0]
        cwes = []

        if vuln.cwe_ids is not None:
            cwes = [int(x.replace("CWE-", "")) for x in vuln.cwe_ids]

        cyclone_vuln = Vulnerability(
            recommendation=vuln.solution,
            cwes=cwes,
            description=vuln.description,
            source=get_vulnerability_source(vuln_id),
        )

        cyclone_vuln.id = vuln_id
        cyclone_vuln.ratings = []
        cyclone_vuln.advisories = get_advisories_from_urls(vuln.urls)
        cyclone_vuln.references = get_references_from_ids(vuln.identifiers, cyclone_vuln.id)

        if vuln.cvss_v3 is not None:
            cvss = CVSS3(vuln.cvss_v3)
            cyclone_vuln.ratings.append(
                cdx.Rating(
                    score=float(cvss.base_score or 0.0),
                    severity=cdx.Severity[cvss.severities()[0].lower()],
                    method=cdx.ScoreMethod.CVSSv3,
                    vector=str(cvss.clean_vector()),
                )
            )
        elif vuln.cvss_v2 is not None:
            cvss = CVSS2(vuln.cvss_v2)
            cyclone_vuln.ratings.append(
                cdx.Rating(
                    score=float(cvss.base_score or 0.0),
                    severity=cdx.Severity[cvss.severities()[0].lower()],
                    method=cdx.ScoreMethod.CVSSv2,
                    vector=cvss.clean_vector(),
                )
            )

        cyclone_vuln.tools = self.scanner_tools()

        return cyclone_vuln

    @staticmethod
    def _get_cache_dir() -> Path:
        cache = os.getenv("CACHE_DIR")
        return Path(cache) if cache is not None else Path(tempfile.gettempdir())

    def _get_path(self, purl: PackageURL) -> Path:
        """Build a path to the gemnasium path."""
        repo_format = purl.type
        namespace = purl.namespace or ""

        if repo_format == "npm":
            path_slug = Path("npm") / namespace / purl.name
        elif repo_format == "maven":
            path_slug = Path("maven") / namespace / purl.name
        elif repo_format == "golang":
            path_slug = Path("go") / namespace / purl.name
        else:
            path_slug = Path(repo_format) / purl.name

        return self.database_path / path_slug



"""A vulnerability enhancement that gets and sets an EPSS probability and percentile if applicable.

--------------------------------------------------------------------------------
SPDX-FileCopyrightText: Copyright © 2024 Lockheed Martin <open.source@lmco.com>
SPDX-FileName: hopprcop/enhancements/epss.py
SPDX-FileType: SOURCE
SPDX-License-Identifier: MIT
--------------------------------------------------------------------------------
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
--------------------------------------------------------------------------------
"""

from __future__ import annotations

import json

from typing import TYPE_CHECKING

from hoppr import cdx

from hopprcop.enhancements.epss.models import EPSSResult
from hopprcop.utils import api_query
from hopprcop.vulnerability_enhancer import BaseEnhancer


if TYPE_CHECKING:
    from hoppr import Vulnerability


class EpssEnhancer(BaseEnhancer, offline_mode_supported=False):
    """Documents an assessment of impact and exploitability."""

    api_start: str = "https://api.first.org/data/v1/epss?cve="
    chunk_size = 100

    def __init__(self, offline_mode: bool = False, enabled: bool = False):
        self.offline_mode = offline_mode
        self.enabled = enabled

        super().__init__()

    def enhance_vulnerabilities(self, vuln_list: list[Vulnerability]):
        """Apply the EPSS score to a list of vulnerabilities."""
        data_results: dict[str, dict[str, str]] = {}
        id_list = [vuln.id for vuln in vuln_list if vuln.id]

        while id_list:
            chunk, id_list = id_list[: self.chunk_size], id_list[self.chunk_size :]
            self.query_epss_api(",".join(chunk), data_results)

        self.set_epss_score(vuln_list, data_results)

    def query_epss_api(self, query: str, data_results: dict[str, dict[str, str]]):
        """Query the EPSS api.

        Updates data_results with the epss score mapped to the cve id.
        """
        response = api_query(f"{self.api_start}{query}")

        if response.status_code == 200:
            epss_result = EPSSResult(**json.loads(response.content))
            for data in epss_result.data:
                data_results[data.cve] = {"epss": data.epss, "percentile": data.percentile}

    def set_epss_score(self, vuln_list: list[Vulnerability], data_results: dict[str, dict[str, str]]):
        """Set epss score for vulnerabilities.

        Returns a dictionary of vulnerabilities with epss score added if applicable.
        """
        for vuln in [vuln for vuln in vuln_list if vuln.id in data_results]:
            if vuln.id and (data := data_results.get(vuln.id)):
                vuln.ratings = vuln.ratings or []
                prob_rating = cdx.Rating(
                    score=float(data.get("epss", "")),
                    method=cdx.ScoreMethod.other,
                    source=cdx.VulnerabilitySource(name="EPSS-Probability", url=self.api_start + vuln.id),
                )
                percentile_rating = cdx.Rating(
                    score=float(data.get("percentile", "")),
                    method=cdx.ScoreMethod.other,
                    source=cdx.VulnerabilitySource(name="EPSS-Percentile", url=self.api_start + vuln.id),
                )
                vuln.ratings.append(prob_rating)
                vuln.ratings.append(percentile_rating)



"""Gitlab dependency scanning data model."""

from __future__ import annotations

from typing import Annotated, Literal

from pydantic import BaseModel, Extra, Field


class Message(BaseModel):
    """Communication intended for the initiator of a scan."""

    level: Annotated[
        Literal["info", "warn", "fatal"],
        Field(
            description=(
                "Describes the severity of the communication. Use info to communicate"
                " normal scan behaviour; warn to communicate a potentially recoverable"
                " problem, or a partial error; fatal to communicate an issue that"
                " causes the scan to halt."
            ),
            examples=["info"],
        ),
    ]
    value: Annotated[
        str,
        Field(
            description="The message to communicate.",
            examples=["Permission denied, scanning aborted"],
            min_length=1,
        ),
    ]


class Option(BaseModel):
    """A configuration option used for this scan."""

    name: Annotated[
        str,
        Field(
            description="The configuration option name.",
            examples=[
                "DAST_FF_ENABLE_BAS",
                "DOCKER_TLS_CERTDIR",
                "DS_MAX_DEPTH",
                "SECURE_LOG_LEVEL",
            ],
            max_length=255,
            min_length=1,
        ),
    ]
    source: Annotated[
        Literal["argument", "file", "env_variable", "other"] | None,
        Field(description="The source of this option."),
    ] = None
    value: Annotated[
        bool | int | str | None,
        Field(
            description="The value used for this scan.",
            examples=[True, 2, None, "fatal", ""],
        ),
    ]


class Vendor(BaseModel):
    """The vendor/maintainer of the analyzer."""

    name: Annotated[
        str,
        Field(description="The name of the vendor.", examples=["GitLab"], min_length=1),
    ]


class Analyzer(BaseModel):
    """Object defining the analyzer used to perform the scan. Analyzers typically delegate to an underlying scanner to run the scan."""

    id: Annotated[
        str,
        Field(
            description="Unique id that identifies the analyzer.",
            examples=["gitlab-dast"],
            min_length=1,
        ),
    ]
    name: Annotated[
        str,
        Field(
            description=("A human readable value that identifies the analyzer, not required to be unique."),
            examples=["GitLab DAST"],
            min_length=1,
        ),
    ]
    url: Annotated[
        str | None,
        Field(
            description="A link to more information about the analyzer.",
            examples=["https://docs.gitlab.com/ee/user/application_security/dast"],
            regex="^https?://.+",
        ),
    ] = None
    vendor: Annotated[Vendor, Field(description="The vendor/maintainer of the analyzer.")]
    version: Annotated[
        str,
        Field(description="The version of the analyzer.", examples=["1.0.2"], min_length=1),
    ]


class Scanner(BaseModel):
    """Object defining the scanner used to perform the scan."""

    id: Annotated[
        str,
        Field(
            description="Unique id that identifies the scanner.",
            examples=["my-sast-scanner"],
            min_length=1,
        ),
    ]
    name: Annotated[
        str,
        Field(
            description=("A human readable value that identifies the scanner, not required to be unique."),
            examples=["My SAST Scanner"],
            min_length=1,
        ),
    ]
    url: Annotated[
        str | None,
        Field(
            description="A link to more information about the scanner.",
            examples=["https://scanner.url"],
        ),
    ] = None
    version: Annotated[
        str,
        Field(description="The version of the scanner.", examples=["1.0.2"], min_length=1),
    ]
    vendor: Annotated[Vendor, Field(description="The vendor/maintainer of the scanner.")]


class PrimaryIdentifier(BaseModel):
    """Primary ID model for Gitlab Dependency Scanning."""

    type: Annotated[
        str,
        Field(
            description=("for example, cve, cwe, osvdb, usn, or an analyzer-dependent type such as gemnasium)."),
            min_length=1,
        ),
    ]
    name: Annotated[str, Field(description="Human-readable name of the identifier.", min_length=1)]
    url: Annotated[
        str | None,
        Field(
            description="URL of the identifier's documentation.",
            regex="^(https?|ftp)://.+",
        ),
    ] = None
    value: Annotated[
        str,
        Field(description="Value of the identifier, for matching purpose.", min_length=1),
    ]


class Scan(BaseModel):
    """Scan model for Gitlab Dependency Scanning."""

    end_time: Annotated[
        str,
        Field(
            description=("ISO8601 UTC value with format yyyy-mm-ddThh:mm:ss, representing when the scan finished."),
            examples=["2020-01-28T03:26:02"],
            regex="^\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}$",
        ),
    ]
    messages: list[Message] | None = None
    options: list[Option] | None = None
    analyzer: Annotated[
        Analyzer,
        Field(
            description=(
                "Object defining the analyzer used to perform the scan. Analyzers"
                " typically delegate to an underlying scanner to run the scan."
            )
        ),
    ]
    scanner: Annotated[
        Scanner,
        Field(description="Object defining the scanner used to perform the scan."),
    ]
    start_time: Annotated[
        str,
        Field(
            description=("ISO8601 UTC value with format yyyy-mm-ddThh:mm:ss, representing when the scan started."),
            examples=["2020-02-14T16:01:59"],
            regex="^\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}$",
        ),
    ]
    status: Annotated[Literal["success", "failure"], Field(description="Result of the scan.")]
    type: Annotated[Literal["dependency_scanning"], Field(description="Type of the scan.")]
    primary_identifiers: Annotated[
        list[PrimaryIdentifier] | None,
        Field(
            description=(
                "An unordered array containing an exhaustive list of primary"
                " identifiers for which the analyzer may return results"
            )
        ),
    ] = None


class Identifier(BaseModel):
    """Identifier model for Gitlab Dependency Scanning."""

    type: Annotated[
        str,
        Field(
            description=("for example, cve, cwe, osvdb, usn, or an analyzer-dependent type such as gemnasium)."),
            min_length=1,
        ),
    ]
    name: Annotated[str, Field(description="Human-readable name of the identifier.", min_length=1)]
    url: Annotated[
        str | None,
        Field(
            description="URL of the identifier's documentation.",
            regex="^(https?|ftp)://.+",
        ),
    ] = None
    value: Annotated[
        str,
        Field(description="Value of the identifier, for matching purpose.", min_length=1),
    ]


class CvssVectors(BaseModel):
    """CVSS Vectors model for Gitlab Dependency Scanning."""

    vendor: Annotated[str | None, Field(min_length=1)] = "unknown"
    vector: Annotated[
        str,
        Field(
            max_length=128,
            min_length=16,
            regex="^((AV:[NAL]|AC:[LMH]|Au:[MSN]|[CIA]:[NPC]|E:(U|POC|F|H|ND)|RL:(OF|TF|W|U|ND)|RC:(UC|UR|C|ND)|CDP:(N|L|LM|MH|H|ND)|TD:(N|L|M|H|ND)|[CIA]R:(L|M|H|ND))/)*(AV:[NAL]|AC:[LMH]|Au:[MSN]|[CIA]:[NPC]|E:(U|POC|F|H|ND)|RL:(OF|TF|W|U|ND)|RC:(UC|UR|C|ND)|CDP:(N|L|LM|MH|H|ND)|TD:(N|L|M|H|ND)|[CIA]R:(L|M|H|ND))$",
        ),
    ]


class CvssVectors3x(BaseModel):
    """Alternate CVSS Vectors model for Gitlab Dependency Scanning."""

    vendor: Annotated[str | None, Field(min_length=1)] = "unknown"
    vector: Annotated[
        str,
        Field(
            max_length=128,
            min_length=32,
            regex="^CVSS:3[.][01]/((AV:[NALP]|AC:[LH]|PR:[NLH]|UI:[NR]|S:[UC]|[CIA]:[NLH]|E:[XUPFH]|RL:[XOTWU]|RC:[XURC]|[CIA]R:[XLMH]|MAV:[XNALP]|MAC:[XLH]|MPR:[XNLH]|MUI:[XNR]|MS:[XUC]|M[CIA]:[XNLH])/)*(AV:[NALP]|AC:[LH]|PR:[NLH]|UI:[NR]|S:[UC]|[CIA]:[NLH]|E:[XUPFH]|RL:[XOTWU]|RC:[XURC]|[CIA]R:[XLMH]|MAV:[XNALP]|MAC:[XLH]|MPR:[XNLH]|MUI:[XNR]|MS:[XUC]|M[CIA]:[XNLH])$",
        ),
    ]


class Link(BaseModel):
    """Link model for Gitlab Dependency Scanning."""

    name: Annotated[str | None, Field(description="Name of the vulnerability details link.")] = None
    url: Annotated[
        str,
        Field(
            description="URL of the vulnerability details document.",
            regex="^(https?|ftp)://.+",
        ),
    ]


class Signature(BaseModel):
    """A calculated tracking signature value and metadata."""

    algorithm: Annotated[str, Field(description="The algorithm used to generate the signature.")]
    value: Annotated[str, Field(description="The result of this signature algorithm.")]


class Item(BaseModel):
    """An item that should be tracked using source-specific tracking methods."""

    file: Annotated[
        str | None,
        Field(description="Path to the file where the vulnerability is located."),
    ] = None
    start_line: Annotated[
        float | None,
        Field(description="The first line of the file that includes the vulnerability."),
    ] = None
    end_line: Annotated[
        float | None,
        Field(description="The last line of the file that includes the vulnerability."),
    ] = None
    signatures: Annotated[
        list[Signature],
        Field(
            description=("An array of calculated tracking signatures for this tracking item."),
            min_items=1,
        ),
    ]


class Tracking(BaseModel):
    """Declares that a series of items should be tracked using source-specific tracking methods."""

    type: Annotated[
        str | None,
        Field(description="Each tracking type must declare its own type."),
    ] = None
    items: list[Item]


class Flag(BaseModel):
    """Informational flags identified and assigned to a vulnerability."""

    type: Annotated[
        Literal["flagged-as-likely-false-positive"],
        Field(description="Result of the scan."),
    ]
    origin: Annotated[str, Field(description="Tool that issued the flag.", min_length=1)]
    description: Annotated[str, Field(description="What the flag is about.", min_length=1)]


class Package(BaseModel):
    """Provides information on the package where the vulnerability is located."""

    name: Annotated[
        str,
        Field(description="Name of the package where the vulnerability is located."),
    ]


class DependencyPath(BaseModel):
    """Dependency Path model for Gitlab Dependency Scanning."""

    iid: Annotated[
        float,
        Field(description=("ID that is unique in the scope of a parent object, and specific to the resource type.")),
    ]


class Dependency(BaseModel):
    """Describes the dependency of a project where the vulnerability is located."""

    package: Annotated[
        Package,
        Field(description=("Provides information on the package where the vulnerability is located.")),
    ]
    version: Annotated[str, Field(description="Version of the vulnerable package.")]
    iid: Annotated[
        float | None,
        Field(description=("ID that identifies the dependency in the scope of a dependency file.")),
    ] = None
    direct: Annotated[
        bool | None,
        Field(description=("Tells whether this is a direct, top-level dependency of the scanned project.")),
    ] = None
    dependency_path: Annotated[
        list[DependencyPath] | None,
        Field(
            description=(
                "Ancestors of the dependency, starting from a direct project"
                " dependency, and ending with an immediate parent of the dependency."
                " The dependency itself is excluded from the path. Direct dependencies"
                " have no path."
            )
        ),
    ] = None


class Location(BaseModel):
    """Identifies the vulnerability's location."""

    file: Annotated[
        str,
        Field(
            description=("Path to the manifest or lock file where the dependency is declared (such as yarn.lock)."),
            min_length=1,
        ),
    ]
    dependency: Annotated[
        Dependency,
        Field(description=("Describes the dependency of a project where the vulnerability is located.")),
    ]


class Fix(BaseModel):
    """Fix model for Gitlab Dependency Scanning."""

    id: Annotated[
        str,
        Field(
            description=("Unique identifier of the vulnerability. This is recommended to be a UUID."),
            examples=["642735a5-1425-428d-8d4e-3c854885a3c9"],
            min_length=1,
        ),
    ]


class Remediation(BaseModel):
    """Remediation model for Gitlab Dependency Scanning."""

    fixes: Annotated[
        list[Fix],
        Field(
            description=("An array of strings that represent references to vulnerabilities fixed by this remediation.")
        ),
    ]
    summary: Annotated[
        str,
        Field(
            description="An overview of how the vulnerabilities were fixed.",
            min_length=1,
        ),
    ]
    diff: Annotated[
        str,
        Field(
            description=("A base64-encoded remediation code diff, compatible with git apply."),
            min_length=1,
        ),
    ]


class DependencyFile(BaseModel):
    """Depnedency File model for Gitlab Dependency Scanning."""

    path: Annotated[str, Field(min_length=1)]
    package_manager: Annotated[str, Field(min_length=1)]
    dependencies: list[Dependency]


class NamedField(BaseModel):
    """Named Field model for Gitlab Dependency Scanning."""

    name: str
    description: str | None = None


class Text(BaseModel):
    """Raw text."""

    type: Literal["text"] = "text"
    value: str


class Url(BaseModel):
    """A single URL."""

    type: Literal["url"] = "url"
    text: str | None = None
    href: Annotated[str, Field(examples=["http://mysite.com"], min_length=1)]


class Code(BaseModel):
    """A codeblock."""

    type: Literal["code"] = "code"
    value: str
    lang: Annotated[str | None, Field(description="A programming language")] = None


class Value(BaseModel):
    """A field that can store a range of types of value."""

    type: Literal["value"] = "value"
    value: float | str | bool


class Diff(BaseModel):
    """A diff."""

    type: Literal["diff"] = "diff"
    before: str
    after: str


class Markdown(BaseModel):
    """GitLab flavoured markdown, see https://docs.gitlab.com/ee/user/markdown.html."""

    type: Literal["markdown"] = "markdown"
    value: Annotated[
        str,
        Field(
            examples=[
                "Here is markdown `inline code` #1 [test](gitlab.com)\n\n![GitLab"
                " Logo](https://about.gitlab.com/images/press/logo/preview/gitlab-logo-white-preview.png)"
            ]
        ),
    ]


class Commit(BaseModel):
    """A commit/tag/branch within the GitLab project."""

    type: Literal["commit"] = "commit"
    value: Annotated[str, Field(description="The commit SHA", min_length=1)]


class FileLocation(BaseModel):
    """A location within a file in the project."""

    type: Literal["file-location"] = "file-location"
    file_name: Annotated[str, Field(min_length=1)]
    line_start: int
    line_end: int | None = None


class ModuleLocation(BaseModel):
    """A location within a binary module of the form module+relative_offset."""

    type: Literal["module-location"] = "module-location"
    module_name: Annotated[str, Field(examples=["compiled_binary"], min_length=1)]
    offset: Annotated[int, Field(examples=[100])]


class Vulnerability(BaseModel):
    """Describes the vulnerability using GitLab Flavored Markdown."""

    id: Annotated[
        str,
        Field(
            description=("Unique identifier of the vulnerability. This is recommended to be a UUID."),
            examples=["642735a5-1425-428d-8d4e-3c854885a3c9"],
            min_length=1,
        ),
    ]
    name: Annotated[
        str | None,
        Field(
            description=("The name of the vulnerability. This must not include the finding's specific information."),
            max_length=255,
        ),
    ] = None
    description: Annotated[
        str | None,
        Field(
            description="A long text section describing the vulnerability more fully.",
            max_length=1048576,
        ),
    ] = None
    severity: Annotated[
        Literal["Info", "Unknown", "Low", "Medium", "High", "Critical"] | None,
        Field(
            description=(
                "How much the vulnerability impacts the software. Possible values are"
                " Info, Unknown, Low, Medium, High, or Critical. Note that some"
                " analyzers may not report all these possible values."
            )
        ),
    ] = None
    solution: Annotated[
        str | None,
        Field(description="Explanation of how to fix the vulnerability.", max_length=7000),
    ] = None
    identifiers: Annotated[
        list[Identifier],
        Field(
            description=(
                "An ordered array of references that identify a vulnerability on"
                " internal or external databases. The first identifier is the Primary"
                " Identifier, which has special meaning."
            ),
            min_items=1,
        ),
    ]
    cvss_vectors: Annotated[
        list[CvssVectors | CvssVectors3x] | None,
        Field(
            description=(
                "An ordered array of CVSS vectors, each issued by a vendor to rate the"
                " vulnerability. The first item in the array is used as the primary"
                " CVSS vector, and is used to filter and sort the vulnerability."
            ),
            max_items=10,
            min_items=1,
        ),
    ] = None
    links: Annotated[
        list[Link] | None,
        Field(
            description=(
                "An array of references to external documentation or articles that describe the vulnerability."
            )
        ),
    ] = None
    details: NamedFieldMapping | None = None
    tracking: Annotated[
        Tracking | None,
        Field(description=("Describes how this vulnerability should be tracked as the project changes.")),
    ] = None
    flags: Annotated[
        list[Flag] | None,
        Field(description="Flags that can be attached to vulnerabilities."),
    ] = None
    location: Annotated[Location, Field(description="Identifies the vulnerability's location.")]


class ReportFormatForGitlabDependencyScanning(BaseModel):
    """This schema provides the the report format for Dependency Scanning analyzers (https://docs.gitlab.com/ee/user/application_security/dependency_scanning)."""

    class Config:
        extra = Extra.allow

    scan: Scan
    schema_: Annotated[
        str | None,
        Field(
            alias="schema",
            description="URI pointing to the validating security report schema.",
            regex="^https?://.+",
        ),
    ] = None
    version: Annotated[
        str,
        Field(
            description="The version of the schema to which the JSON report conforms.",
            regex="^[0-9]+\\.[0-9]+\\.[0-9]+$",
        ),
    ]
    vulnerabilities: Annotated[list[Vulnerability], Field(description="Array of vulnerability objects.")]
    remediations: Annotated[
        list[Remediation] | None,
        Field(
            description=(
                "An array of objects containing information on available remediations, along with patch diffs to apply."
            )
        ),
    ] = None
    dependency_files: Annotated[
        list[DependencyFile],
        Field(description="List of dependency files identified in the project."),
    ]


class NamedList(BaseModel):
    """An object with named and typed fields."""

    type: Literal["named-list"] = "named-list"
    items: NamedFieldMapping | None = None


class ListModel(BaseModel):
    """A list of typed fields."""

    type: Literal["list"] = "list"
    items: list[DetailType]


class Table(BaseModel):
    """A table of typed fields."""

    type: Literal["table"] = "table"
    header: list[DetailType] | None = None
    rows: list[list[DetailType]]


class NamedFieldNamedListItem(NamedField, NamedList):
    """A combination model of NamedField and NamedList to support the definition of named_list/properties/items."""


class NamedFieldListItem(NamedField, ListModel):
    """A combination model of NamedField and ListModel to support the definition of named_list/properties/items."""


class NamedFieldTableItem(NamedField, Table):
    """A combination model of NamedField and Table to support the definition of named_list/properties/items."""


class NamedFieldTextItem(NamedField, Text):
    """A combination model of NamedField and Text to support the definition of named_list/properties/items."""


class NamedFieldUrlItem(NamedField, Url):
    """A combination model of NamedField and Url to support the definition of named_list/properties/items."""


class NamedFieldCodeItem(NamedField, Code):
    """A combination model of NamedField and Code to support the definition of named_list/properties/items."""


class NamedFieldValueItem(NamedField, Value):
    """A combination model of NamedField and Value to support the definition of named_list/properties/items."""


class NamedFieldDiffItem(NamedField, Diff):
    """A combination model of NamedField and Diff to support the definition of named_list/properties/items."""


class NamedFieldMarkdownItem(NamedField, Markdown):
    """A combination model of NamedField and Markdown to support the definition of named_list/properties/items."""


class NamedFieldCommitItem(NamedField, Commit):
    """A combination model of NamedField and Commit to support the definition of named_list/properties/items."""


class NamedFieldFileLocationItem(NamedField, FileLocation):
    """A combination model of NamedField and FileLocation to support the definition of named_list/properties/items."""


class NamedFieldModuleLocationItem(NamedField, ModuleLocation):
    """A combination model of NamedField and ModuleLocation to support the definition of named_list/properties/items."""


DetailType = Annotated[
    NamedList
    | ListModel
    | Table
    | Text
    | Url
    | Code
    | Value
    | Diff
    | Markdown
    | Commit
    | FileLocation
    | ModuleLocation,
    Field(discriminator="type"),
]

NamedFieldMapping = Annotated[
    dict[
        str,
        NamedFieldListItem
        | NamedFieldTableItem
        | NamedFieldTextItem
        | NamedFieldUrlItem
        | NamedFieldCodeItem
        | NamedFieldValueItem
        | NamedFieldDiffItem
        | NamedFieldMarkdownItem
        | NamedFieldCommitItem
        | NamedFieldFileLocationItem
        | NamedFieldModuleLocationItem,
    ],
    Field(default={}),
]

Vulnerability.update_forward_refs()
NamedList.update_forward_refs()
NamedField.update_forward_refs()
NamedFieldNamedListItem.update_forward_refs()
NamedFieldListItem.update_forward_refs()
NamedFieldTableItem.update_forward_refs()
NamedFieldTextItem.update_forward_refs()
NamedFieldUrlItem.update_forward_refs()
NamedFieldCodeItem.update_forward_refs()
NamedFieldValueItem.update_forward_refs()
NamedFieldDiffItem.update_forward_refs()
NamedFieldMarkdownItem.update_forward_refs()
NamedFieldCommitItem.update_forward_refs()
NamedFieldFileLocationItem.update_forward_refs()
NamedFieldModuleLocationItem.update_forward_refs()



# ruff: noqa
"""Exceptions that can be thrown from API interaction."""
#
# Copyright 2022-Present Sonatype Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations


class OssIndexException(Exception):
    """Base exception which all exceptions raised by this library extend."""


class AccessDeniedException(OssIndexException):
    """Raised if supplied credentials for Oss Index are invalid."""


class RateLimitException(OssIndexException):
    """Raised if oss index returns a 429 too many requests exception."""
